# -*- coding: utf-8 -*-
"""
@author: Ke Zhai (zhaike@cs.umd.edu)

Implements collapsed Gibbs sampling for the linear-Gaussian infinite latent feature model (IBP).
"""

import numpy, scipy;
import math, random;
from .gs import GibbsSampling;

# We will be taking log(0) = -Inf, so turn off this warning
numpy.seterr(divide='ignore')

class CollapsedGibbsSampling(GibbsSampling):
    import scipy.stats;

    """
    @param data: a NxD NumPy data matrix
    @param alpha: IBP hyper parameter
    @param sigma_x: standard derivation of the noise on data, often referred as sigma_n as well
    @param sigma_a: standard derivation of the feature, often referred as sigma_f as well
    @param initializ_Z: seeded Z matrix
    """
    def _initialize(self, data, alpha=1.0, sigma_a=1.0, sigma_x=1.0, A_prior=None, initial_Z=None):
        super(CollapsedGibbsSampling, self)._initialize(self.center_data(data), alpha, sigma_a, sigma_x, initial_Z);

        if A_prior == None:
            self._A_prior = numpy.zeros((1, self._D));
        else:
            self._A_prior = A_prior;

        assert(self._A_prior.shape == (1, self._D));

        # compute matrix M
        self._M = self.compute_M();
        self._log_det_M = numpy.log(numpy.linalg.det(self._M));

        assert(numpy.abs(numpy.log(numpy.linalg.det(self._M)) - self._log_det_M) < 0.000000001)

    """
    sample the corpus to train the parameters
    """
    def sample(self, iteration, directory="../../output/tmp-output/"):
        import os, shutil
        if os.path.exists(directory):
            shutil.rmtree(directory);

        assert(self._Z.shape == (self._N, self._K));
        assert(self._X.shape == (self._N, self._D));

        #sample the total data
        for iter in xrange(iteration):
            # sample every object
            order = numpy.random.permutation(self._N);
            for (object_counter, object_index) in enumerate(order):
                if object_counter > 0 and object_counter % 100 == 0:
                    print("sampling in progress %2d%%" % (100 * object_counter / self._N));

                # compute M_i
                ziM = numpy.dot(self._Z[[object_index], :], self._M);
                ziMzi = numpy.dot(ziM, self._Z[[object_index], :].transpose());
                M_i = self._M - numpy.dot(ziM.transpose(), ziM) / (ziMzi - 1);
                log_det_M_i = self._log_det_M - numpy.log(1 - ziMzi);

                # sample Z_n
                singleton_features = self.sample_Zn(object_index, M_i, log_det_M_i);

                if self._metropolis_hastings_k_new:
                    # sample K_new using metropolis hasting
                    self.metropolis_hastings_K_new(object_index, singleton_features, M_i, log_det_M_i);

            self._A = self.map_estimate_A();

            if self._alpha_hyper_parameter != None:
                self._alpha = self.sample_alpha();

            if self._sigma_x_hyper_parameter != None:
                self._sigma_x = self.sample_sigma_x(self._sigma_x_hyper_parameter);

            if self._sigma_a_hyper_parameter != None:
                self._sigma_a = self.sample_sigma_a(self._sigma_a_hyper_parameter);

            print("iteration: %i\tK: %i\tlikelihood: %f" % (iter, self._K, self.log_likelihood_model()));
            print("alpha: %f\tsigma_a: %f\tsigma_x: %f" % (self._alpha, self._sigma_a, self._sigma_x));

            if (iter + 1) % self._snapshot_interval == 0:
                self.export_snapshot(directory, iter + 1);

    """
    @param object_index: an int data type, indicates the object index (row index) of Z we want to sample
    """
    def sample_Zn(self, object_index, M_i, log_det_M_i):
        assert(type(object_index) == int or type(object_index) == numpy.int32 or type(object_index) == numpy.int64);

        # calculate initial feature possess counts
        m = self._Z.sum(axis=0);

        # remove this data point from m vector
        new_m = (m - self._Z[object_index, :]).astype(numpy.float);

        # compute the log probability of p(Znk=0 | Z_nk) and p(Znk=1 | Z_nk)
        log_prob_z1 = numpy.log(new_m / self._N);
        log_prob_z0 = numpy.log(1 - new_m / self._N);

        # find all singleton features possessed by current object
        singleton_features = [nk for nk in range(self._K) if self._Z[object_index, nk] != 0 and new_m[nk] == 0];
        non_singleton_features = [nk for nk in range(self._K) if nk not in singleton_features]
        order = numpy.random.permutation(self._K);

        for (feature_counter, feature_index) in enumerate(order):
            if feature_index in non_singleton_features:
                old_Znk = self._Z[object_index, feature_index];

                # compute the log likelihood when Znk=1
                self._Z[object_index, feature_index] = 1;
                if old_Znk == 0:
                    ziMi = numpy.dot(self._Z[[object_index], :], M_i);
                    ziMizi = numpy.dot(ziMi, self._Z[[object_index], :].transpose());
                    M_tmp_1 = M_i - numpy.dot(ziMi.transpose(), ziMi) / (ziMizi + 1);
                    log_det_M_tmp_1 = log_det_M_i - numpy.log(ziMizi + 1);

                    #assert(numpy.abs(log_det_M_tmp_1 - numpy.log(numpy.linalg.det(M_tmp_1))) < 0.000000001);
                else:
                    M_tmp_1 = self._M;
                    log_det_M_tmp_1 = self._log_det_M;

                prob_z1 = self.log_likelihood_X(M_tmp_1, log_det_M_tmp_1);
                # add in prior
                prob_z1 += log_prob_z1[feature_index];

                # compute the log likelihood when Znk=0
                self._Z[object_index, feature_index] = 0;
                if old_Znk == 1:
                    ziMi = numpy.dot(self._Z[[object_index], :], M_i);
                    ziMizi = numpy.dot(ziMi, self._Z[[object_index], :].transpose());
                    M_tmp_0 = M_i - numpy.dot(ziMi.transpose(), ziMi) / (ziMizi + 1);
                    log_det_M_tmp_0 = log_det_M_i - numpy.log(ziMizi + 1);
                else:
                    M_tmp_0 = self._M;
                    log_det_M_tmp_0 = self._log_det_M;

                prob_z0 = self.log_likelihood_X(M_tmp_0, log_det_M_tmp_0);
                # add in prior
                prob_z0 += log_prob_z0[feature_index];

                #print "propose znk to 0", numpy.exp(prob_z1-prob_z0);
                Znk_is_0 = 1 / (1 + numpy.exp(prob_z1 - prob_z0));
                #print "znk is 0 with prob", Znk_is_0
                if random.random() < Znk_is_0:
                    self._Z[object_index, feature_index] = 0;
                    self._M = M_tmp_0;
                    self._log_det_M = log_det_M_tmp_0;
                else:
                    self._Z[object_index, feature_index] = 1;
                    self._M = M_tmp_1;
                    self._log_det_M = log_det_M_tmp_1;

        return singleton_features;

    """
    sample K_new using metropolis hastings algorithm
    """
    def metropolis_hastings_K_new(self, object_index, singleton_features, M_i, log_det_M_i):
        # sample K_new from the metropolis hastings proposal distribution, i.e., a poisson distribution with mean \frac{\alpha}{N}
        K_temp = scipy.stats.poisson.rvs(self._alpha / self._N);

        if K_temp <= 0 and len(singleton_features) <= 0:
            return False;

        # compute the probability of using old features
        prob_old = self.log_likelihood_X();

        # construct Z_new
        #Z_new = self._Z[:, [k for k in range(self._K) if k not in singleton_features]];
        Z_new = numpy.hstack((self._Z, numpy.zeros((self._N, K_temp))));
        Z_new[[object_index], [xrange(-K_temp, 0)]] = 1;
        Z_new[[object_index], singleton_features] = 0;

        # construct M_new
        M_i_new = numpy.vstack((numpy.hstack((M_i, numpy.zeros((self._K, K_temp)))), numpy.hstack((numpy.zeros((K_temp, self._K)), (self._sigma_a / self._sigma_x) ** 2 * numpy.eye(K_temp)))));
        log_det_M_i_new = log_det_M_i + 2 * K_temp * numpy.log(self._sigma_a / self._sigma_x);
        ziMi = numpy.dot(Z_new[[object_index], :], M_i_new);
        ziMizi = numpy.dot(ziMi, Z_new[[object_index], :].transpose());
        M_new = M_i_new - numpy.dot(ziMi.transpose(), ziMi) / (ziMizi + 1);
        log_det_M_new = log_det_M_i_new - numpy.log(ziMizi + 1);
        K_new = self._K + K_temp;
        #assert(numpy.abs(log_det_M_new - numpy.log(numpy.linalg.det(M_new))) < 0.000000001);

        # compute the probability of using new features
        prob_new = self.log_likelihood_X(M_new, log_det_M_new, Z_new);

        # compute the probability of generating new features
        accept_new = 1 / (1 + numpy.exp(prob_old - prob_new));

        # if we accept the proposal, we will replace old A and Z matrices
        if random.random() < accept_new:
            self._Z = Z_new;
            self._K = K_new;
            self.regularize_matrices();
            return True;

        return False;

    """
    remove the empty column in matrix Z and the corresponding feature in A
    """
    def regularize_matrices(self):
        assert(self._Z.shape == (self._N, self._K));
        Z_sum = numpy.sum(self._Z, axis=0);
        assert(len(Z_sum) == self._K);
        indices = numpy.nonzero(Z_sum == 0);

        self._Z = self._Z[:, [k for k in range(self._K) if k not in indices]];
        self._K = self._Z.shape[1];
        assert(self._Z.shape == (self._N, self._K));

        # compute matrix M
        self._M = self.compute_M();
        self._log_det_M = numpy.log(numpy.linalg.det(self._M));

    """
    compute the log-likelihood of the data X
    @param X: a 2-D numpy array
    @param Z: a 2-D numpy boolean array
    @param A: a 2-D numpy array, integrate A out if it is set to None
    """
    def log_likelihood_X(self, M=None, log_det_M=None, Z=None):
        if M == None:
            M = self._M;
            if log_det_M == None:
                log_det_M = numpy.log(numpy.linalg.det(M));
            else:
                log_det_M = self._log_det_M;

        if Z == None:
            Z = self._Z;

        assert(self._X.shape[0] == Z.shape[0]);
        (N, D) = self._X.shape;
        (N, K) = Z.shape;
        assert(M.shape == (K, K));

        # we are collapsing A out, i.e., compute the log likelihood p(X | Z)
        # be careful that M passed in should include the inverse.
        log_likelihood = numpy.eye(N) - numpy.dot(numpy.dot(Z, M), Z.transpose());
        log_likelihood = -0.5 / (self._sigma_x ** 2) * numpy.trace(numpy.dot(numpy.dot(self._X.transpose(), log_likelihood), self._X));
        log_likelihood -= D * (N - K) * numpy.log(self._sigma_x) + K * D * numpy.log(self._sigma_a);
        log_likelihood += 0.5 * D * log_det_M;
        log_likelihood -= 0.5 * N * D * numpy.log(2 * numpy.pi);

        return log_likelihood

    """
    compute the log-likelihood of the model
    """
    def log_likelihood_model(self):
        #print self.log_likelihood_X(self._X, self._Z, self._A_mean), self.log_likelihood_A(), self.log_likelihood_Z();
        return self.log_likelihood_X() + self.log_likelihood_Z();

    """
    sample noise variances, i.e., sigma_x
    """
    def sample_sigma_x(self, sigma_x_hyper_parameter):
        return self.sample_sigma(self._sigma_x_hyper_parameter, self._X - numpy.dot(self._Z, self._A));

    """
    sample feature variance, i.e., sigma_a
    """
    def sample_sigma_a(self, sigma_a_hyper_parameter):
        return self.sample_sigma(self._sigma_a_hyper_parameter, self._A);

"""
run IBP on the synthetic 'cambridge bars' dataset, used in the original paper.
"""
if __name__ == '__main__':
    import scipy.io;

    # load the data from the matrix
    mat_vals = scipy.io.loadmat('../../data/cambridge-bars/block_image_set.mat');
    true_weights = mat_vals['trueWeights']
    features = mat_vals['features']
    data = mat_vals['data']

    # set up the hyper-parameter for sampling alpha
    alpha_hyper_parameter = (1., 1.);
    # set up the hyper-parameter for sampling sigma_x
    sigma_x_hyper_parameter = (1., 1.);
    # set up the hyper-parameter for sampling sigma_a
    sigma_a_hyper_parameter = (1., 1.);

    features = features.astype(numpy.int);

    # initialize the model
    ibp = CollapsedGibbsSampling(alpha_hyper_parameter, sigma_x_hyper_parameter, sigma_a_hyper_parameter, True);

    ibp._initialize(data[1:100, :], 0.5, 0.2, 0.5);
    #ibp._initialize(data[0:1000, :], 1.0, 1.0, 1.0, None, features[0:1000, :]);

    #print ibp._Z, "\n", ibp._A_mean
    ibp.sample(30);

    print(ibp._Z.sum(axis=0))

    # If matplotlib is installed, plot ground truth vs learned factors
    import matplotlib.pyplot as P
    from util.scaled_image import scaledimage

    # Intensity plots of
    # -ground truth factor-feature weights (top)
    # -learned factor-feature weights (bottom)
    K = max(len(true_weights), len(ibp._A))
    (fig, subaxes) = P.subplots(2, K)
    for sa in subaxes.flatten():
        sa.set_visible(False)
    fig.suptitle('Ground truth (top) vs learned factors (bottom)')
    for (idx, trueFactor) in enumerate(true_weights):
        ax = subaxes[0, idx]
        ax.set_visible(True)
        scaledimage(trueFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
    for (idx, learnedFactor) in enumerate(ibp._A):
        ax = subaxes[1, idx]
        scaledimage(learnedFactor.reshape(6, 6),
                    pixwidth=3, ax=ax)
        ax.set_visible(True)
    P.show()



from pymake import ExpSpace, ExpTensor, Corpus, ExpDesign, ExpGroup


class Netw2(ExpDesign):


    data_net_all = Corpus(['manufacturing', 'fb_uc','blogs', 'emaileu', 'propro', 'euroroad', 'generator7', 'generator12', 'generator10', 'generator4'])
    net_all = data_net_all + Corpus(['clique6', 'BA'])


    # compare perplexity and rox curve from those baseline.
    compare_scvb = ExpTensor (
        corpus        = ['clique6', 'BA'],
        model         = ['immsb_cgs', 'ilfm_cgs', 'rescal', 'immsb_cvb'],
        N             = 200,
        K             = 6,
        iterations    = 30,
        hyper         = 'auto',
        testset_ratio = 20,
        homo = 0,
        mask = 'unbalanced',

        _data_type    = 'networks',
        _refdir       = 'debug_scvb' ,
        _format       = '{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}',
        _csv_typo     = '_iteration time_it _entropy _entropy_t _K _alpha _gmma alpha_mean delta_mean alpha_var delta_var'
    )
    compare_scvb_m = ExpGroup(compare_scvb, model=['immsb_cgs', 'immsb_cvb'])
    cvb = ExpGroup(compare_scvb, model='immsb_cvb')

    scvb = ExpTensor (
        corpus        = ['BA'],
        model         = 'immsb_scvb',
        N             = 200,
        chunk         = 'adaptative_1',
        K             = 6,
        iterations    = 3,
        hyper         = 'auto',
        testset_ratio = 20,
        #chi_a = 1,
        #tau_a = 42,
        kappa_a = 0.75,
        #chi_b = 42,
        #tau_b = 300,
        #kappa_b = 0.9,

        homo = 0,
        mask = 'unbalanced',

        _data_type    = 'networks',
        _refdir       = 'debug_scvb' ,
        _format       = '{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}',
        _csv_typo     = '_iteration time_it _entropy _entropy_t _K _chi_a _tau_a _kappa_a _chi_b _tau_b _kappa_b _elbo _roc'
    )
    scvb_t = ExpGroup(scvb, _refdir='debug_')

    scvb_chi = ExpGroup(scvb, chi_a=1, tau_a=42, kappa_a=[0.6, 0.7, 0.9],
                        chi_b=42, tau_b=300, kappa_b=[0.6, 0.7, 0.9],
                        _format='{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}'
                       )

    scvb_chi2 = ExpGroup(scvb, chi_a=[10], tau_a=[100], kappa_a=[0.6],
                        chi_b=[10], tau_b=[500], kappa_b=[0.9],
                        _format='{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}'
                       )

    scvb_chi_2 = ExpGroup(scvb, chi_a=[1, 10], tau_a=[42, 100, 500], kappa_a=[0.6],
                        chi_b=[1, 10], tau_b=[42, 100, 500], kappa_b=[0.6, 0.7, 0.9],
                        _format='{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}'
                       )


    #
    #
    #
    # * iter1 : don"t set masked, and gamma is not symmetric
    #
    # * iter2 : don"t set masked, and gamma is symmetric
    #
    #
    #


    scvb1_chi_a = ExpTensor (
        corpus        = ['blogs', 'manufacturing', 'generator7', 'generator10'],
        model         = 'immsb_scvb',
        N             = 'all',
        chunk         = 'adaptative_1',
        K             = 6,
        iterations    = 1,
        hyper         = 'auto',
        testset_ratio = 20,
        chi_a = [0.5, 1, 2, 10],
        tau_a = [0.5, 1, 2, 16, 64, 256, 1024],
        kappa_a = [0.51, 0.45, 1],
        #homo = 0,
        #mask = 'unbalanced',

        _data_type    = 'networks',
        _refdir       = 'debug_scvb_chia_i1' ,
        _format       = '{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}-{_name}-{_id}',
        _csv_typo     = '_iteration time_it _entropy _entropy_t _K _chi_a _tau_a _kappa_a _chi_b _tau_b _kappa_b'
    )

    scvb1_chi_b = ExpTensor (
        corpus        = ['blogs', 'manufacturing', 'generator7', 'generator10'],
        model         = 'immsb_scvb',
        N             = 'all',
        chunk         = 'adaptative_1',
        K             = 6,
        iterations    = 1,
        hyper         = 'auto',
        testset_ratio = 20,
        chi_b = [0.5, 1, 2, 10],
        tau_b = [0.5, 1, 2, 16, 64, 256, 1024],
        kappa_b = [0.51, 0.45, 1],
        #homo = 0,
        #mask = 'unbalanced',

        _data_type    = 'networks',
        _refdir       = 'debug_scvb_chib_i1' ,
        _format       = '{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}-{_name}-{_id}',
        _csv_typo     = '_iteration time_it _entropy _entropy_t _K _chi_a _tau_a _kappa_a _chi_b _tau_b _kappa_b'
    )


    # Noel expe
    noel = ExpGroup([scvb, compare_scvb], N='all', corpus=data_net_all, mask=['balanced', 'unbalanced'], _refdir='noel')

    noel_cvb = ExpGroup(cvb, N='all', corpus=data_net_all, mask=['balanced', 'unbalanced'], _refdir='noel')
    noel_scvb = ExpGroup(scvb, N='all', corpus=data_net_all, mask=['balanced', 'unbalanced'], _refdir='noel')
    noel_scvb_ada = ExpGroup(noel_scvb, chunk=['adaptative_0.1', 'adaptative_0.5', 'adaptative_1', 'adaptative_10'])

    noel_mmsb = ExpGroup([scvb, compare_scvb_m], N='all', corpus=data_net_all,
                    mask=['balanced', 'unbalanced'], _refdir='noel')

    compare_scvb2 = ExpGroup(compare_scvb, N='all', corpus=data_net_all, iterations=100,
                    mask=['balanced', 'unbalanced'], _refdir='noel2')

    noel3 = ExpGroup(scvb_chi2, N='all', chunk=['adaptative_0.1', 'adaptative_1'], corpus=data_net_all, mask=['unbalanced'], _refdir='noel3')


    # cvb debug
    pd = ExpGroup(compare_scvb, iterations=150, model='immsb_cvb', _repeat='debug_cvb',
                  N='all', corpus=data_net_all,
                  mask=['unbalanced'], _refdir='noel2')

    pd2n = ExpGroup(compare_scvb, iterations=150, model='immsb_cvb', _repeat='debug_cvb_2n',
                  N='all', corpus=data_net_all,
                  mask=['unbalanced'], _refdir='noel2')







from pymake import ExpSpace, ExpTensor, Corpus, ExpDesign, ExpGroup


class Levy(ExpDesign):

    #refdir=gapv1 is old likelihood + sampling r/p.sampling
    #rocw1: new likelihood + sampling r/p
    #rocw2: newlikelihood + mixed samling/VB r/p
    #rocw3: new likelihood + mixed samling/VB r/p and N_Phi added. (partial some correction and inversion sample/expectation)
    #rocw4: new likelihood + samling r/p and

    _alias = {'ml.iwmmsb_scvb3' : 'WMMSB',
              'ml.iwmmsb_scvb3_auto' : 'WMMSB-bg',
              'ml.immsb_scvb3' : 'MMSB',
              'ml.sbm_gt' : 'SBM',
              'ml.wsbm_gt' : 'WSBM',
              'link-dynamic-simplewiki': 'wiki-link',
              'munmun_digg_reply': 'digg-reply',
              'slashdot-threads': 'slashdot',
             }

    net_old = Corpus(['manufacturing', 'fb_uc','blogs', 'emaileu', 'propro', 'euroroad', 'generator7', 'generator12', 'generator10', 'generator4'])
    net_gt = Corpus(['astro-ph', 'hep-th', 'netscience', 'cond-mat']) # all undirected
    net_random = Corpus(['clique6', 'BA'])
    net_test = Corpus(['manufacturing', 'fb_uc', 'netscience'])
    net_large = Corpus(['link-dynamic-simplewiki', 'enron', 'foldoc'])
    net_w = Corpus(['manufacturing', 'fb_uc']) + net_gt
    # 'actor-collaboration' # Too big?
    net_w2 = Corpus(['slashdot-threads','prosper-loans', 'munmun_digg_reply', 'moreno_names' ])
    # manufacturing
    net_final = Corpus(['fb_uc', 'manufacturing','hep-th', 'link-dynamic-simplewiki', 'enron', 'slashdot-threads', 'prosper-loans', 'munmun_digg_reply', 'moreno_names', 'astro-ph'])
    net_all = net_old + net_gt + net_large

    #
    # Poisson Point process :
    # * stationarity / ergodicity of p(d_i) ?
    # * Erny theorem (characterization by void probabilities ?
    # * Inference ? Gamma Process ?
    # * Sparsity ?

    wmmsb = ExpTensor (
        corpus        = ['BA'],
        model         = 'iwmmsb_scvb',
        N             = 'all',
        chunk         = 'adaptative_1',
        K             = 6,
        iterations    = 3,
        hyper         = 'auto',
        testset_ratio = 20,

        delta = [[2,2]],

        chi_a=10, tau_a=100, kappa_a=0.6,
        chi_b=10, tau_b=500, kappa_b=0.9,

        homo = 0,
        mask = 'unbalanced',

        _data_format = 'w',
        _data_type    = 'networks',
        _refdir       = 'debug_scvb' ,
        _format='{corpus}_{model}_{N}_{K}_{iterations}_{hyper}_{homo}_{mask}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}',
        _csv_typo     = '_iteration time_it _entropy _entropy_t _K _chi_a _tau_a _kappa_a _chi_b _tau_b _kappa_b _elbo _roc _pr'
    )
    noelw3 = ExpGroup(wmmsb, N='all', chunk=['adaptative_0.1', 'adaptative_1'], corpus=net_old, mask=['unbalanced'], _refdir='noel3')



    warm = ExpTensor(
        corpus        = ['manufacturing'],
        model         = 'iwmmsb_scvb3',
        N             = 'all',
        K             = 10,
        hyper         = 'auto',
        homo = 0,
        testset_ratio = 20,
        validset_ratio = 10,


        # Sampling
        chunk         = 'stratify',
        sampling_coverage = 0.42,
        #chi_a=10, tau_a=100, kappa_a=0.6,
        #chi_b=10, tau_b=500, kappa_b=0.9,
        chi_a=1, tau_a=1024, kappa_a=0.5,
        chi_b=1, tau_b=1024, kappa_b=0.5,
        zeros_set_prob = 1/2,
        zeros_set_len = 50,

        #delta = [[1, 1]],
        #delta = [[0.5, 10]],
        delta = 'auto',

        fig_xaxis = [('_observed_pt', 'visited edges')],
        fig_legend = 4,
        legend_size = 12,
        #ticks_size = 20,
        title_size = 20,

        driver = 'gt', # graph-tool driver

        _data_type    = 'networks',
        _refdir       = 'debug_scvb3',
        _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}',
        _csv_typo     = '_observed_pt time_it _entropy _K _chi_a _tau_a _kappa_a _chi_b _tau_b _kappa_b _roc _wsim _pr',
    )

    # Sparse sampling
    warm_sparse_chunk = ExpGroup(warm, sampling_coverage=1, chunk='sparse')

    # Sampling sensiblity | hyper-delta sensibility
    warm_sampling = ExpGroup(warm, delta=[[1,1],[0.5,10],[10,0.5]],
                             zeros_set_prob = [1/2, 1/3, 1/4],  zeros_set_len=[10, 50])

    arm_sampling = ExpGroup(warm_sampling, delta='auto', model='immsb_scvb3')

    sbm_base = ExpGroup(warm, model=['sbm_gt', 'wsbm_gt', 'rescal_als'],
                        zeros_set_prob=None, zeros_set_len=None, delta=None,
                        _csv_typo='time_it _entropy _K _roc _wsim _pr')
    wsbm2_base = ExpGroup(sbm_base, model = ['wsm_g', 'wsbm2_gt'])
    wsbm_base = ExpGroup(sbm_base, model = ['wsbm_gt'])

    # Compare sensibility
    eta_b = ExpGroup([arm_sampling], _refdir='eta', corpus=net_w,
                   zeros_set_prob=[1/2, 1/4])
    eta_w = ExpGroup([warm_sampling], _refdir='eta', corpus=net_w,
                   zeros_set_prob=[1/2, 1/4])
    eta = ExpGroup([eta_b, eta_w])

    eta_sbm = ExpGroup(sbm_base, _refdir='eta', corpus=net_w)


    # test/visu
    warm_visu = ExpGroup(warm, delta=[[1,1],'auto'],
                         zeros_set_prob = [1/2],  zeros_set_len=[10])

    # Best selection visu (eta)
    best_mmsb = ExpGroup(eta_b, zeros_set_prob=1/4, zeros_set_len=50, delta='auto')
    best_wmmsb = ExpGroup(eta_w, zeros_set_prob=1/4, zeros_set_len=50, delta=[[0.5,10]])
    best_scvb = ExpGroup([best_mmsb, best_wmmsb])
    eta_best = ExpGroup([best_scvb, eta_sbm])


    # Corrected zero sampling
    eta2_base = ExpGroup(warm, testset_ratio=20, _refdir='roc5', corpus=net_w) # roc5§roc5_N
    eta2_sbm = ExpGroup(eta2_base, model=['sbm_gt', 'wsbm_gt', 'rescal_als'],
                        zeros_set_prob=None, zeros_set_len=None, delta=None,
                        _csv_typo='time_it _entropy _K _roc _wsim _pr')

    eta2_b = ExpGroup(eta2_base, model="immsb_scvb3", zeros_set_prob=1/2, zeros_set_len=10, delta='auto')
    eta2_b10 = eta2_b
    eta2_b50 = ExpGroup(eta2_base, model="immsb_scvb3", zeros_set_prob=1/2, zeros_set_len=50, delta='auto')

    eta0_w = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=10, delta=[[10, 0.5]])
    eta1_w = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=10, delta=[[1, 1]])
    eta2_w = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=10, delta=[[0.5, 10]])
    eta3_w = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=10, delta=[[0.1, 10]])
    eta0_w50 = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=50, delta=[[10, 0.5]])
    eta1_w50 = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=50, delta=[[1, 1]])
    eta2_w50 = ExpGroup(eta2_base, model="iwmmsb_scvb3", zeros_set_prob=1/2, zeros_set_len=50, delta=[[0.5, 10]])

    eta2a_w = ExpGroup(eta2_base, model="iwmmsb_scvb3_auto", zeros_set_prob=1/2, zeros_set_len=10, delta='auto', _model="ml.iwmmsb_scvb3")
    eta2a_w10 = eta2a_w
    eta2a_w50 = ExpGroup(eta2_base, model="iwmmsb_scvb3_auto", zeros_set_prob=1/2, zeros_set_len=50, delta='auto', _model="ml.iwmmsb_scvb3")

    eta4_full = ExpGroup([eta2_b50, eta2_w50, eta2a_w50]) # weighte are squared

    roc_visu_sbm = ExpGroup(sbm_base, corpus=net_w, testset_ratio=20, model=['sbm_gt', 'wsbm_gt'], _refdir='roc5')
    roc_visu_sbm_full = ExpGroup(sbm_base, corpus=net_w, testset_ratio=20, model=['rescal_als', 'sbm_gt', 'wsbm_gt'], _refdir='roc5')
    roc_visu_full = ExpGroup([eta4_full, roc_visu_sbm_full])
    roc_visu_final10 = ExpGroup([eta2_b, eta2_w, eta2a_w, roc_visu_sbm])
    roc_visu_final50 = ExpGroup([eta2_b50, eta2_w50, eta2a_w50, roc_visu_sbm])
    roc_visu_final = roc_visu_final50


    roc_visu_final2 = ExpGroup([eta2_b50, eta2_w50, eta2a_w50], _refdir="roc5v")
    roc_visu_final2 = ExpGroup([roc_visu_final2, roc_visu_sbm])
    roc_visu_final2_full = ExpGroup([roc_visu_final2, roc_visu_sbm_full])

    roc_visu_final3 = ExpGroup([eta2_b50, eta1_w50, eta2a_w50], _refdir="roc5v")
    roc_visu_final3 = ExpGroup([roc_visu_final3, roc_visu_sbm])

    online_roc = ExpGroup(roc_visu_final3 , training_ratio=[1, 5,10,20,30, 50, 100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )

    online_roc_ = ExpGroup([eta2_b50, eta2a_w50, roc_visu_sbm] , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )


    online_roc_sbm = ExpGroup(roc_visu_sbm , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )

    online_roc_mmsb = ExpGroup([eta2_b50, eta1_w50, eta2a_w50] , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )

    online_roc_w = ExpGroup([eta1_w50, eta2a_w50, wsbm_base] , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )

    online_roc_wm = ExpGroup([eta1_w50, eta2a_w50] , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                            )

    online_roc_wsbm = ExpGroup([wsbm_base] , training_ratio=[1, 5,10,20,30,50,100], _refdir='online1w', corpus=net_final,
                          _seed='corpus', testset_ratio=20,
                          _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}--{training_ratio}',
                         )

    roc_w = ExpGroup([eta0_w, eta1_w, eta2_w, eta0_w50, eta1_w50, eta2_w50]) # squared weight
    roc_b = ExpGroup(eta2_base, model="immsb_scvb3", zeros_set_prob=1/2, zeros_set_len=[10, 50], delta='auto')

    conv_w = ExpGroup([eta0_w50, eta1_w50, eta2_w50, eta2a_w50], corpus=['astro-ph', 'enron', 'munmun_digg_reply'])

    param1 = ExpGroup(eta2a_w50,  c0=[0.5, 1,10, 100], r0=[0.1, 0.5, 1])
    gap = ExpGroup([param1], _refdir='gap_hyper', testset_ratio=20, corpus=net_w,
                   ce=[100], eps=[1e-6],
                   _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}-{c0}-{r0}-{ce}-{eps}',
                  )


    gap_visu = ExpGroup([eta2a_w50], _refdir='gap_hyper', testset_ratio=20, corpus=net_w,
                    ce=[1, 10, 100], eps=[1e-5, 1e-6, 1e-7],
                    c0=10, r0=1,
                    _format='{corpus}_{model}_{N}_{K}_{hyper}_{homo}_{testset_ratio}_{chunk}_{chi_a}-{tau_a}-{kappa_a}_{chi_b}-{tau_b}-{kappa_b}_{delta}_{zeros_set_len}_{zeros_set_prob}-{c0}-{r0}-{ce}-{eps}',
                  )





# -*- coding: utf-8 -*-


from ml.model.hdp.mmsb import GibbsRun, Likelihood, ZSampler, NP_CGS
from ml.model.hdp.hdp import MSampler, BetaSampler

from ml.model.hdp.mmsb import CGS, ZSamplerParametric

# @idem than ilda_cgs
class immsb_cgs(GibbsRun):
    def __init__(self, expe, frontend):

        hyperparams = expe.get('hyperparams', {})
        delta = hyperparams.get('delta',1)
        alpha = hyperparams.get('alpha',1)
        gmma = hyperparams.get('gmma',1)

        hyper = expe.hyper
        assortativity = expe.get('homo')
        hyper_prior = expe.get('hyper_prior') # HDP hyper optimization
        K = expe.K

        likelihood = Likelihood(delta, frontend, assortativity=assortativity)

        if frontend:
            likelihood._symmetric = frontend.is_symmetric()

            # Nonparametric case
            zsampler = ZSampler(alpha, likelihood, K_init=K)
            msampler = MSampler(zsampler)
            betasampler = BetaSampler(gmma, msampler)
            jointsampler = NP_CGS(zsampler, msampler, betasampler,
                                  hyper=hyper, hyper_prior=hyper_prior)

            self.s = jointsampler
        else:
            likelihood._symmetric = None


        super().__init__(expe, frontend)
        self.update_hyper(hyperparams)


class mmsb_cgs(GibbsRun):
    def __init__(self, expe, frontend):

        hyperparams = expe.get('hyperparams', {})
        delta = hyperparams.get('delta',1)
        alpha = hyperparams.get('alpha',1)
        gmma = hyperparams.get('gmma',1)

        hyper = expe.hyper
        assortativity = expe.get('homo')
        hyper_prior = expe.get('hyper_prior') # HDP hyper optimization
        K = expe.K

        likelihood = Likelihood(delta, frontend, assortativity=assortativity)

        if frontend:
            likelihood._symmetric = frontend.is_symmetric()

            # Parametric case
            jointsampler = CGS(ZSamplerParametric(alpha, likelihood, K))

            self.s = jointsampler
        else:
            likelihood._symmetric = None

        super().__init__(expe, frontend)
        self.update_hyper(hyperparams)




import numpy as np
#np.set_printoptions(threshold='nan')

from pymake import ExpeFormat
from pymake.manager import ModelManager, FrontendManager

from pymake.util.utils import colored



''' Build and Stats for Text data '''


class Text(ExpeFormat):

    _default_expe = dict(
        _spec = 'data_text_all',
        _data_type = 'text'
    )

    def __call__(self):
        return self.stats()

    def stats(self):
        ''' Show data stats '''
        raise NotImplementedError



import sys, os
from datetime import datetime
from collections import defaultdict
import logging
import hashlib
import json
from string import Template

import numpy as np
import scipy as sp

from builtins import input
basestring = (str, bytes)


try:
    from terminal import colorize
    colored = lambda *x : str(colorize(x[0], x[1]))
except ImportError:
    lgg = logging.getLogger('root')
    lgg.debug("needs `terminal' module for colors printing")
    colored = lambda *x : x[0]


#from itertools import cycle
class Cycle(object):
    def __init__(self, seq):
        self.seq = seq
        self.it = np.nditer([seq])
    def next(self):
        return self.__next__()
    def __next__(self):
        try:
            return next(self.it).item()
        except StopIteration:
            self.it.reset()
            # Exception on nditer when seq is empty (infinite recursivity)
            return self.next()

    def reset(self):
        return self.it.reset()

    def copy(self):
        return self.__class__(self.seq)

def get_dest_opt_filled(parser):
    ''' Return the {dest} name of the options filled in the command line

        Parameters
        ----------
        parser : ArgParser

        Returns
        -------
        set of string
    '''

    opts_in = [opt for opt in sys.argv if opt.startswith('-') and opt not in ['-vv','-vvv']]
    opt2dest_dict = dict( (opt, act.dest) for act in parser._get_optional_actions() for opt in act.option_strings )
    dests_in = set([opt2dest_dict[opt] for opt in opts_in])
    return dests_in

# Assign new values to an array according to a map list
def set_v_to(a, map):
    new_a = a.copy()
    for k, c in dict(map).iteritems():
        new_a[a==k] = c

    return new_a

# Re-order the confusion matrix in order to map the cluster (columns) to the best (classes) according to purity
# One class by topics !
# It modify confu and map in-place
# Return: list of tuple that map topic -> class
import sys
sys.setrecursionlimit(10000)
def map_class2cluster_from_confusion(confu, map=None, cpt=0, minmax='max'):
    assert(confu.shape[0] == confu.shape[1])

    if minmax == 'max':
        obj_f = np.argmax
    else:
        obj_f = np.argmin

    if len(confu) -1  == cpt:
        # Recursive stop condition
        return sorted(map)
    if map is None:
        confu = confu.copy()
        map = [ (i,i) for i in range(len(confu)) ]
        #map = np.array(map)

    #K = confu.shape[0]
    #C = confu.shape[1]
    previous_assigned = [i[1] for i in map[:cpt]]
    c_l = obj_f(np.delete(confu[cpt], previous_assigned))
    # Get the right id of the class
    for j in sorted(previous_assigned):
        # rectify c_l depending on which class where already assigning
        if c_l >= j:
            c_l += 1
        else:
            break
    m_l = confu[cpt, c_l]
    # Get the right id of the topic
    c_c = obj_f(confu[cpt:,c_l]) + cpt
    m_c = confu[c_c, c_l]
    if m_c > m_l:
        # Move the line corresponding to the max for this class to the top
        confu[[cpt, c_c], :] = confu[[c_c, cpt], :]
        map[cpt], map[c_c] = map[c_c], map[cpt] # Doesn't work if it's a numpy array
        return map_class2cluster_from_confusion(confu, map, cpt)
    else:
        # Map topic 1 to class c_l and return confu - topic 1 and class c_l
        map[cpt] = (map[cpt][0], c_l)
        cpt += 1
        return map_class2cluster_from_confusion(confu, map, cpt)


def drop_zeros(a_list):
    #return [i for i in a_list if i>0]
    return filter(lambda x: x != 0, a_list)

def nxG(y):
    import networkx as nx
    if type(y) is np.ndarray:
        if (y == y.T).all():
            # Undirected Graph
            typeG = nx.Graph()
        else:
            # Directed Graph
            typeG = nx.DiGraph()
        G = nx.from_numpy_matrix(y, create_using=typeG)
    else:
        G = y
    return G

#
#
# Common/Utils
#
#

def retrieve_git_info():
    git_branch = subprocess.check_output(['git', 'rev-parse','--abbrev-ref' ,'HEAD']).strip().decode()
    git_hash = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode()

    return {'git_branch':git_branch, 'git_hash':git_hash}

def hash_objects(obj, algo='md5'):
    """ Return a list of hash of the input object """
    hashalgo = getattr(hashlib, algo)

    """ Return a hash of the input """
    if isinstance(obj, (np.ndarray, list, tuple)):
        # array of int
        hashed_obj = hashalgo(np.asarray(obj).tobytes()).hexdigest()
    elif isinstance(obj, str):
        hashed_obj = hashalgo(obj.encode("utf-8")).hexdigest()
    elif isinstance(obj, dict):
        hashed_obj = hashalgo(json.dumps(obj, sort_keys=True).encode('utf8')).hexdigest()
    else:
        raise TypeError('Type of object unashable: %s' % (type(obj)))

    return hashed_obj

def ask_sure_exit(question):

    while True:
        a = input(question+' ').lower()
        if a in ('yes', 'y'):
            break
        elif a in ('no', 'n'):
            exit(2)
        else:
            print("Enter either [y|n]")

def make_path(f):
    bdir = os.path.dirname(f)
    if not os.path.exists(bdir) and bdir:
        os.makedirs(bdir)
    #fn = os.path.basename(bdir)
    #if not os.path.exists(fn) and fn:
    #    open(fn, 'a').close()
    return bdir



def Now():
    return  datetime.now()
def nowDiff(last):
    return datetime.now() - last
def ellapsed_time(text, since):
    current = datetime.now()
    delta = current - since
    print(text + ' : %s' % (delta))
    return current

def tail(filename, n_lines):
    _tail = []
    for i, line in enumerate(reverse_readline(filename)):
        if i == n_lines:
            break
        _tail.append(line)
    return _tail[::-1]
#import mmap
#def tail(filename, nlines):
#    """Returns last n lines from the filename. No exception handling"""
#    size = os.path.getsize(filename)
#    with open(filename, "rb") as f:
#        # for Windows the mmap parameters are different
#        fm = mmap.mmap(f.fileno(), 0, mmap.MAP_SHARED, mmap.PROT_READ)
#        try:
#            for i in range(size - 1, -1, -1):
#                if fm[i] == '\n':
#                    nlines -= 1
#                    if nlines == -1:
#                        break
#            return fm[i + 1 if i else 0:].splitlines()
#        finally:
#            pass
#

def reverse_readline(filename, buf_size=8192):
    """a generator that returns the lines of a file in reverse order"""
    with open(filename) as fh:
        segment = None
        offset = 0
        fh.seek(0, os.SEEK_END)
        file_size = remaining_size = fh.tell()
        while remaining_size > 0:
            offset = min(file_size, offset + buf_size)
            fh.seek(file_size - offset)
            buffer = fh.read(min(remaining_size, buf_size))
            remaining_size -= buf_size
            lines = buffer.split('\n')
            # the first line of the buffer is probably not a complete line so
            # we'll save it and append it to the last line of the next buffer
            # we read
            if segment is not None:
                # if the previous chunk starts right from the beginning of line
                # do not concact the segment to the last line of new chunk
                # instead, yield the segment first
                if buffer[-1] is not '\n':
                    lines[-1] += segment
                else:
                    yield segment
            segment = lines[0]
            for index in range(len(lines) - 1, 0, -1):
                if len(lines[index]):
                    yield lines[index]
        # Don't yield None if the file was empty
        if segment is not None:
            yield segment


class defaultdict2(defaultdict):
	def __missing__(self, key):
		if self.default_factory is None:
			raise KeyError( key )
		else:
			ret = self[key] = self.default_factory(key)
			return ret






#
# Deprecated
#

import numpy, scipy
import scipy.special

"""
@author: Ke Zhai (zhaike@cs.umd.edu)

Implements variational bayes for the hierarchical Dirichlet process (HDP).
"""
class VariationalBayes(object):
    """
    """
    def __init__(self,
                 snapshot_interval=10,
                 finite_mode=True,
                 _lambda=None,
                 model_likelihood_threshold=0.00001,
                 global_maximum_iteration=100):
        self._global_maximum_iteration = global_maximum_iteration
        self._finite_mode = finite_mode
        assert(_lambda == None or type(_lambda) == tuple)
        self._lambda = _lambda
        self._model_likelihood_threshold = model_likelihood_threshold

        self._snapshot_interval = snapshot_interval

    """
    @param data:
    take note: words are not terms, they are repeatable and thus might be not unique
    """
    def _initialize(self, data, K=5):
        # initialize the total number of topics.
        self._K = K

        # initialize the documents, key by the document path, value by a list of non-stop and tokenized words, with duplication.
        self._data = data

        # initialize the size of the collection, i.e., total number of documents.
        self._D = len(self._data)

        # initialize the vocabulary, i.e. a list of distinct tokens.
        self._vocab = []
        for token_list in data.values():
            self._vocab += token_list
        self._vocab = list(set(self._vocab))

        # initialize the size of the vocabulary, i.e. total number of distinct tokens.
        self._V = len(self._vocab)

        # initialize a V-by-K matrix phi, valued at random.
        self._phi = numpy.random.random((self._V, self._K))
        assert(self._phi.shape == (self._V, self._K))

        # initialize a (K-1)-by-2 matrix gamma, valued at random.
        self._gamma = numpy.random.random((self._K, 2))
        assert(self._gamma.shape == (self._K, 2))

        # initialize a K-by-2 matrix beta, valued at random.
        self._tau = numpy.random.random((self._K, 2))
        assert(self._tau.shape == (self._K, 2))

    """
    """
    def update_phi(self):
        nu_sum_n = numpy.sum(self._nu, axis=0)
        assert(len(nu_sum_n) == self._K)
        nu_sum_k = numpy.sum(self._nu, axis=1)
        assert(len(nu_sum_k) == self._N)
        for k in numpy.random.permutation(xrange(self._K)):
            # we assume the covariance matrix is a diagonal matrix and we only store the diagonal terms
            self._phi_cov[k, :] = 1. / (1. / (self._sigma_a * self._sigma_a) + nu_sum_n[k] / (self._sigma_x * self._sigma_x))

            tmp_nu = numpy.delete(self._nu, k, 1)
            assert(tmp_nu.shape == (self._N, self._K - 1))
            tmp_phi_mean = numpy.delete(self._phi_mean, k, 0)
            assert(tmp_phi_mean.shape == (self._K - 1, self._D))
            phi_mean = numpy.dot(self._X.transpose(), self._nu[:, k][:, numpy.newaxis]).transpose()
            for n in xrange(self._N):
                phi_mean -= self._nu[n, k] * numpy.sum(tmp_nu[n, :] * tmp_phi_mean.transpose(), axis=1)[:, numpy.newaxis].transpose()
            assert(phi_mean.shape == (1, self._D))
            phi_mean /= self._sigma_x * self._sigma_x
            self._phi_mean[k, :] = phi_mean * self._phi_cov[k, :]

        return

    """
    """
    def update_gamma(self):
        self._gamma[0, :] = 1 + numpy.sum(self._phi, axis=0)

        var_theta_constant = self.compute_var_theta_constant()
        for k in numpy.random.permutation(xrange(self._K)):
            tmp_nu = numpy.delete(self._nu, k, 1)
            assert(tmp_nu.shape == (self._N, self._K - 1))
            tmp_phi_mean = numpy.delete(self._phi_mean, k, 0)
            assert(tmp_phi_mean.shape == (self._K - 1, self._D))

            for n in numpy.random.permutation(xrange(self._N)):
                x_nu_phi = self._X[n, :][numpy.newaxis, :] - numpy.sum(tmp_nu[n, :] * tmp_phi_mean.transpose(), axis=1)[:, numpy.newaxis].transpose()
                assert(x_nu_phi.shape == (1, self._D))
                var_theta = var_theta_constant[k] + 1. / (self._sigma_x * self._sigma_x) * numpy.dot(self._phi_mean[k, :][numpy.newaxis, :], x_nu_phi.transpose())
                self._nu[n, k] = 1. / (1. + numpy.exp(-var_theta))

        return

    """
    """
    def update_tau(self):
        sum_nu = numpy.sum(self._nu, axis=0)
        assert(len(sum_nu) == self._K)
        if self._finite_mode:
            assert(len(sum_nu) == self._K)
            self._tau[0, :] = self._alpha / self._K + sum_nu
            self._tau[1, :] = 1. + self._N - sum_nu
        else:
            N_minus_sum_nu = self._N - sum_nu
            for k in xrange(self._K):
                psi_tau = scipy.special.psi(self._tau)
                assert(psi_tau.shape == ( self._K,2))
                #assert(psi_tau.shape == (2, self._K))

                psi_sum_tau = scipy.special.psi(numpy.sum(self._tau, axis=0))
                assert(len(psi_sum_tau) == self._K)

                psi_tau0_cumsum = numpy.hstack([0, numpy.cumsum(psi_tau[0, :-1])])
                assert(len(psi_tau0_cumsum) == self._K)

                psi_sum_cumsum = numpy.cumsum(psi_sum_tau)
                assert(len(psi_sum_cumsum) == self._K)

                exponent = psi_tau[1, :] + psi_tau0_cumsum - psi_sum_cumsum
                unnormalized = numpy.exp(exponent - numpy.max(exponent))
                assert(len(unnormalized) == self._K)

                qs = numpy.zeros((self._K, self._K))
                for m in xrange(k, self._K):
                    qs[m, 0:m + 1] = unnormalized[0:m + 1] / numpy.sum(unnormalized[0:m + 1])

                self._tau[0, k] = numpy.sum(sum_nu[k:self._K]) + numpy.dot(N_minus_sum_nu[k + 1:self._K], numpy.sum(qs[k + 1:self._K, k + 1:self._K], axis=1)) + self._alpha
                self._tau[1, k] = numpy.dot(N_minus_sum_nu[k:self._K], qs[k:self._K, k]) + 1
        return

    """
    """
    def compute_var_theta_constant(self):
        var_theta_constant = -0.5 / (self._sigma_x * self._sigma_x) * ((numpy.sum(self._phi_cov, axis=1) + numpy.sum(self._phi_mean * self._phi_mean, axis=1)))
        assert(len(var_theta_constant) == self._K)
        if self._finite_mode:
            var_theta_constant += scipy.special.psi(self._tau[0, :]) - scipy.special.psi(self._tau[1, :])
        else:
            for k in xrange(self._K):
                var_theta_constant[k] += numpy.sum(scipy.special.psi(self._tau[0, 0:k]) - scipy.special.psi(numpy.sum(self._tau[:, 0:k])))
                var_theta_constant[k] -= self.compute_expected_pzk0_qjensen(k)
        return var_theta_constant

    """
    """
    def compute_expected_pzk0_qjensen(self, k):
        assert(k >= 0 and k < self._K)
        tau = self._tau[:, 0:k + 1]
        assert(tau.shape == (2, k + 1))

        psi_tau = scipy.special.psi(tau)
        assert(psi_tau.shape == (2, k + 1))

        psi_sum_tau = scipy.special.psi(numpy.sum(tau, axis=0))
        assert(len(psi_sum_tau) == k + 1)

        psi_tau0_cumsum = numpy.hstack([0, numpy.cumsum(psi_tau[0, :-1])])
        assert(len(psi_tau0_cumsum) == k + 1)

        psi_sum_cumsum = numpy.cumsum(psi_sum_tau)
        assert(len(psi_sum_cumsum) == k + 1)

        tmp = psi_tau[1, :] + psi_tau0_cumsum - psi_sum_cumsum
        assert(len(tmp) == k + 1)

        q = numpy.exp(tmp - numpy.max(tmp))
        assert(len(q) == k + 1)

        q = q / numpy.sum(q)
        assert(len(q) == k + 1)

        # compute the lower bound
        lower_bound = numpy.sum(q * (tmp - numpy.log(q)))

        return lower_bound

    """
    """
    def learning(self, iteration=0, directory="../../output/tmp-output"):
        if iteration <= 0:
            iteration = self._global_maximum_iteration

        print ("likelihood is", self.velb())

        for i in xrange(iteration):
            self.update_tau()
            self.update_phi()
            self.update_nu()
            print ("likelihood is", self.velb())

        print ("learning finished...")

    """
    compute the variational evidence lower bound (i.e. velb) in log scale
    """
    def velb(self):
        log_likelihood = numpy.zeros(5)

        psi_tau = scipy.special.psi(self._tau)
        assert(psi_tau.shape == ( self._K,2))
        #assert(psi_tau.shape == (2, self._K))
        psi_sum_tau = scipy.special.psi(numpy.sum(self._tau, axis=1)[numpy.newaxis, :])
        assert(psi_sum_tau.shape == (1, self._K))

        if self._finite_mode:
            # compute the probability of feature
            log_likelihood[0] = self._K * numpy.log(self._alpha / self._K) + (self._alpha / self._K - 1.) * numpy.sum(psi_tau[0, :] - psi_sum_tau)
            # compute the probability of feature statistics
            log_likelihood[1] = numpy.sum(self._nu * psi_tau[0, :]) + numpy.sum((1. - self._nu) * psi_tau[1, :]) - self._N * numpy.sum(psi_sum_tau)
        else:
            # compute the probability of feature
            log_likelihood[0] = self._K * numpy.log(self._alpha) + (self._alpha - 1.) * numpy.sum(psi_tau[0, :] - psi_sum_tau)
            # compute the probability of feature statistics
            for k in xrange(self._K):
                log_likelihood[1] += numpy.sum(self._nu[:, k]) * numpy.sum(psi_tau[0, :k + 1] - psi_sum_tau[0, :k + 1])
                log_likelihood[1] += numpy.dot((self._N - numpy.sum(self._nu[:, k])), self.compute_expected_pzk0_qjensen(k))

        # compute the probability of feature distribution
        log_likelihood[2] = -0.5 * self._K * self._D * numpy.log(2 * numpy.pi * self._sigma_a * self._sigma_a)
        log_likelihood[2] -= 0.5 / (self._sigma_a * self._sigma_a) * (numpy.sum(self._phi_cov) + numpy.sum(self._phi_mean * self._phi_mean))

        # compute the probability of data likelihood
        tmp_log_likelihood = numpy.sum(self._X * self._X) - 2 * numpy.sum(self._nu * numpy.dot(self._X, self._phi_mean.transpose()))
        tmp_1 = numpy.dot(numpy.ones((self._N, self._D)), (self._phi_cov + self._phi_mean ** 2).transpose())
        tmp_log_likelihood += numpy.sum(self._nu * tmp_1)
        tmp_1 = numpy.dot(self._nu, self._phi_mean)
        tmp_2 = numpy.sum(numpy.dot(self._nu ** 2, self._phi_mean ** 2))
        tmp_log_likelihood += numpy.sum(tmp_1 * tmp_1) - numpy.sum(tmp_2)

        log_likelihood[3] = -0.5 * self._N * self._D * numpy.log(2 * numpy.pi * self._sigma_x * self._sigma_x)
        log_likelihood[3] -= 0.5 / (self._sigma_x * self._sigma_x) * tmp_log_likelihood

        # entropy of the proposed distribution
        lngamma_tau = scipy.special.gammaln(self._tau)
        assert(lngamma_tau.shape == (2, self._K))
        lngamma_sum_tau = scipy.special.gammaln(numpy.sum(self._tau, axis=0)[numpy.newaxis, :])
        assert(lngamma_sum_tau.shape == (1, self._K))

        # compute the entropy of the distribution
        log_likelihood[4] = numpy.sum(lngamma_tau[0, :] + lngamma_tau[1, :] - lngamma_sum_tau)
        log_likelihood[4] -= numpy.sum((self._tau[0, :] - 1) * psi_tau[0, :] + (self._tau[1, :] - 1) * psi_tau[1, :])
        log_likelihood[4] += numpy.sum((self._tau[0, :] + self._tau[1, :] - 2) * psi_sum_tau)

        assert(numpy.all(self._phi_cov > 0))
        assert(numpy.all(self._nu >= 0) and numpy.all(self._nu <= 1))
        log_likelihood[4] += 0.5 * self._K * self._D * numpy.log(2 * numpy.pi * numpy.e)
        log_likelihood[4] += 0.5 * numpy.sum(numpy.log(self._phi_cov))
        #log_likelihood[4] += 0.5 * numpy.log(numpy.sqrt(numpy.sum(self._phi_cov * self._phi_cov, axis=1)))
        log_likelihood[4] -= numpy.sum(self._nu * numpy.log(self._nu) + (1. - self._nu) * numpy.log(1. - self._nu))

        return numpy.sum(log_likelihood)


"""
some utility functions
"""
def import_monolingual_data(input_file):
    import codecs
    input = codecs.open(input_file, mode="r", encoding="utf-8")

    doc_count = 0
    docs = {}

    for line in input:
        line = line.strip().lower()

        contents = line.split("\t")
        assert(len(contents) == 2)
        docs[int(contents[0])] = [int(item) for item in contents[1].split()]

        doc_count += 1
        if doc_count % 10000 == 0:
            print ("successfully import " + str(doc_count) + " documents...")

    print ("successfully import all documents...")
    return docs
"""
this method convert a corpus from dict(list) to defaultdict(FreqDist) format, similar words are grouped
@return: a defaultdict(FreqDist) data type, first indexed by the document id, then indexed by the unique tokens
@param corpus: a dict data type, indexed by document id, corresponding value is a list of words (not necessarily unique from each other)
"""
def dict_list_2_dict_freqdist(corpus):
    from collections import defaultdict
    from nltk.probability import FreqDist

    docs = defaultdict(dict)

    for doc in corpus.keys():
        docs[doc] = FreqDist(corpus[doc])

    return docs

if __name__ == "__main__":
    from util.vocabulary import Vocabulary, parse_raw_text

    bdir = "../../data/nips12/"
    #data = import_monolingual_data(bdir + "doc.dat")
    data = parse_raw_text(bdir)
    data = data[:50]
    voca = Vocabulary(exclude_stopwords=True)
    corpus = [voca.doc2bow(doc) for doc in data]
    #corpus = np.array([np.random.poisson(1, k) for k in np.random.poisson(50, 20)])
    corpus = dict( (i,j) for i,j in enumerate(corpus) )

    vb = VariationalBayes()
    vb._initialize(corpus)
    vb.learning(iteration=100)



import numpy as np
from scipy.special import psi

from ml.model.modelbase import ModelBase

class MMSB(object):
    """
    Mixed-membership stochastic block models, Airoldi et al., 2008

    """

    def __init__(self, expe, Y):
        """ follows the notations in the original NIPS paper

        :param Y: node by node interaction matrix, row=sources, col=destinations
        :param K: number of mixtures
        :param rho: sparsity parameter
        """
        self.N = int(Y.shape[0])    # number of nodes
        self.K = expe['K']
        self.Y = Y

        self.alpha = np.asarray([1.0 / (i + np.sqrt(self.K)) for i in range(self.K)])
        self.optimize_rho = False
        self.max_iter = expe['iterations']

        #variational parameters
        self.phi = np.random.dirichlet([1]*self.K, size=(self.N, self.N))

        self.gamma = np.random.dirichlet([1]*self.K, size=self.N)

        self.B = np.random.random(size=(self.K,self.K))

        self.rho = (1.-np.sum(self.Y)/(self.N*self.N))  # 1 - density of matrix

    def variational_inference(self, converge_ll_fraction=1e-3):
        """ run variational inference for this model
        maximize the evidence variational lower bound

        :param converge_ll_fraction: terminate variational inference when the fractional change of the lower bound falls below this
        """
        converge = False
        old_ll = 0.
        iteration = 0

        while not converge:
            ll = self.run_e_step()
            ll += self.run_m_step()

            iteration += 1

            #if (old_ll-ll)/ll < converge_ll_fraction or iteration >= self.max_iter:
            if iteration >= self.max_iter:
                converge = True

            print('iteration %d, lower bound %.2f' %(iteration, ll))


    def run_e_step(self):
        """ compute variational expectations
        """
        ll = 0.

        for p in range(self.N):
            for q in range(self.N):
                new_phi = np.zeros(self.K)

                for g in range(self.K):
                    new_phi[g] = np.exp(psi(self.gamma[p,g])-psi(np.sum(self.gamma[p,:]))) * np.prod(( (self.B[g,:]**self.Y[p,q])
                        * ((1.-self.B[g,:])**(1.-self.Y[p,q])) )
                        ** self.phi[q,p,:] )
                self.phi[p,q,:] = new_phi/np.sum(new_phi)

                new_phi = np.zeros(self.K)
                for h in range(self.K):
                    new_phi[h] = np.exp(psi(self.gamma[q,h])-psi(np.sum(self.gamma[q,:]))) * np.prod(( (self.B[:,h]**self.Y[p,q])
                        * ((1.-self.B[:,h])**(1.-self.Y[p,q])) )
                        ** self.phi[p,q,:] )
                self.phi[q,p,:] = new_phi/np.sum(new_phi)

                for k in range(self.K):
                    self.gamma[p,k] = self.alpha[k] + np.sum(self.phi[p,:,k]) + np.sum(self.phi[:,p,k])
                    self.gamma[q,k] = self.alpha[k] + np.sum(self.phi[q,:,k]) + np.sum(self.phi[:,q,k])

        return ll

    def run_m_step(self):
        """ maximize the hyper parameters
        """
        ll = 0.

        self.optimize_alpha()
        self.optimize_B()
        if self.optimize_rho:
            self.update_rho()

        return ll

    def optimize_alpha(self):
        return

    def optimize_B(self):
        for g in range(self.K):
            for h in range(self.K):
                tmp1=0.
                tmp2=0.
                for p in range(self.N):
                    for q in range(self.N):
                        tmp = self.phi[p,q,g]*self.phi[q,p,h]
                        tmp1 += self.Y[p,q]*tmp
                        tmp2 += tmp
                self.B[g,h] = tmp1/tmp2
        return

    def update_rho(self):
        return

class mmsb_vb(ModelBase):

    def __init__(self, expe, frontend):
        super().__init__(expe, frontend)

        self.expe = expe

        self.frontend = frontend
        self.mask = self.frontend.data_ma.mask

    def fit(self):
        data = self.frontend.data_ma
        K = self.expe.K

        model = MMSB(self.expe, data)
        model.variational_inference()

        self._theta = model.gamma
        self._phi = model.B


if __name__ == '__main__':

    """ test with interaction matrix:
    1 1 0 0 0
    1 1 0 0 0
    0 0 1 1 1
    0 0 1 1 1
    0 0 1 1 1
    """
    Y = np.array([[1,1,0,0,0],[1,1,0,0,0],[0,0,1,1,1],[0,0,1,1,1],[0,0,1,1,1]])
    model = MMSB(Y, 2)
    model.variational_inference()

    print(model.gamma)
    print(model.B)






from time import time
import sys
import numpy as np
import scipy as sp
from numpy import ma
import scipy.stats

from pymake.util.utils import defaultdict2
from pymake.util.math import lognormalize, categorical, sorted_perm, adj_to_degree, gem
from ml.model.modelbase import SVB

#import warnings
#warnings.filterwarnings('error')
#warnings.catch_warnings()
##np.seterr(all='print')


class ldpfa_scvb3_np(SVB):
    def __init__(self, expe, frontend):

        if expe['delta'] != 'auto':
            raise ValueError("Delta should be `auto' for wmmsb_scvb_np model...")

        super().__init__(expe, frontend)

class ldpfa_scvb3(SVB):

    _purge = ['_kernel', '_lut_nbinom', '_likelihood']

    def _init_params(self, frontend):
        self.frontend = frontend

        # Save the testdata
        self._max_w = 10
        if hasattr(self.frontend, 'data_test'):
            self.data_test = frontend.data_test_w
            self._max_w = self.data_test[:,2].max()
            self._min_w = self.data_test[:,2].min() or 1

        # Data statistics
        _len = {}
        _len['K'] = self.expe.get('K')
        _len['N'] = frontend.num_nodes()
        _len['E'] = frontend.num_edges()
        _len['nnz'] = frontend.num_nnz()
        #_len['nnz_t'] = frontend.num_nnz_t()
        _len['dims'] = frontend.num_neighbors()
        _len['nnz_ones'] = frontend.num_edges()
        _len['nnzsum'] = frontend.num_nnzsum()
        self._len = _len

        self._K = self._len['K']
        self._is_symmetric = frontend.is_symmetric()

        # Eta init
        self._eta = []
        self._eta_limit = 1e-3
        self._eta_control = -1
        self._eta_count_init = 25
        self._eta_count = self._eta_count_init

        self._init_gradient()

        # Hyperparams
        self.hyper_theta = np.asarray([1.0 / (i + np.sqrt(self._len['K'])) for i in range(self._len['K'])])
        self.hyper_theta /= self.hyper_theta.sum()

        self.hyper_phi = np.asarray(self.expe['delta'])
        hyper_phi = self.expe['delta']


        self._random_ss_init()

        shift = self.expe.get('shift_w')
        #
        # Warn: Instable Part
        #
        if hyper_phi == 'auto':
            K = self._K
            N = self._len['N']
            self._hyper_phi = 'auto'
            #self.c0 = 1
            #self.r0 = 1
            #self.ce = 1
            #self.eps = 1e-3

            self.c0 = float(self.expe.get('c0', 10)) #weights mean
            self.r0 = float(self.expe.get('r0', 0.1))
            self.ce = float(self.expe.get('ce', 10))
            self.eps = float(self.expe.get('eps', 1e-6))

            self.c0_r0 = self.c0*self.r0
            self.ce_eps = self.ce*self.eps
            self.ce_minus_eps = self.ce*(1-self.eps)

            rk = np.ones((K,K)) * self.r0
            pk = np.ones((K,K)) * self.eps
            self.hyper_phi = np.array([rk, (1-pk)/pk])

            self.log.info('Optimizing hyper enabled.')
            #self.hyper_phi = np.array([1,1])
        elif shift:
            a = 10
            b = shift
            self.hyper_phi = np.array([a,b])
        elif len(hyper_phi) == 2:
            self.hyper_phi = np.asarray(hyper_phi)
            self._hyper_phi = 'fix'
        else:
            raise ValueError('hyper parmeter hyper_phi dont understood: %s' % hyper_phi)

    def _init_gradient(self):
        N = self._len['N']
        self._timestep_a = np.zeros(N)
        self.gstep_theta = np.zeros(N)
        self._timestep_b = 0
        self._timestep_c = 0

        self._chi_a = self.expe.get('chi_a', 5)
        self._tau_a = self.expe.get('tau_a', 10)
        self._kappa_a = self.expe.get('kappa_a', 0.9)
        self._chi_b = self.expe.get('chi_b', 1)
        self._tau_b = self.expe.get('tau_b', 100)
        self._kappa_b = self.expe.get('kappa_b', 0.9)

        self._update_gstep_theta(np.arange(len(self._timestep_a)))
        self._update_gstep_phi()
        self._update_gstep_y()


    def _random_ss_init(self):
        ''' Sufficient Statistics Initialization '''
        K = self._len['K']
        N = self._len['N']
        nnz = self._len['nnz']
        nnzsum = self._len['nnzsum']
        dims = self._len['dims']

        self.N_theta_left = (dims[:, None] * np.random.dirichlet([0.5]*K, N))
        self.N_theta_right = (dims[:, None] * np.random.dirichlet([0.5]*K, N))

        if self.expe.get('homo') == 'assortative':
            N_phi_d = np.diag(np.random.dirichlet([0.5]*K)) *nnz*3/4
            N_phi_d1 = np.diag(np.random.dirichlet([0.5]*(K-1)), 1) *nnz*1/8
            if self._is_symmetric:
                N_phi_d1 += np.diag(np.random.dirichlet([0.5]*(K-1)), -1) *nnz*1/8
                du = np.diag(np.ones(K-1), 1)==1
                dl = np.diag(np.ones(K-1), -1)==1
                N_phi_d1[dl] = N_phi_d1[du]
            else:
                N_phi_d1 += np.diag(np.random.dirichlet([0.5]*(K-1)), -1) *nnz*1/8
            N_phi = N_phi_d + N_phi_d1
            self.N_phi = ma.masked_where(N_phi==0, N_phi)

            N_Y_d = np.diag(np.random.dirichlet([0.5]*K)) *nnzsum*3/4
            N_Y_d1 = np.diag(np.random.dirichlet([0.5]*(K-1)), 1) *nnzsum*1/8
            if self._is_symmetric:
                N_Y_d1 += np.diag(np.random.dirichlet([0.5]*(K-1)), -1) *nnzsum*1/8
                du = np.diag(np.ones(K-1), 1)==1
                dl = np.diag(np.ones(K-1), -1)==1
                N_Y_d1[dl] = N_Y_d1[du]
            else:
                N_Y_d1 += np.diag(np.random.dirichlet([0.5]*(K-1)), -1) *nnzsum*1/8
            N_Y = N_Y_d + N_Y_d1
            self.N_Y = ma.masked_where(N_Y==0, N_Y)

        else:
            self.N_phi = np.random.dirichlet([0.5]*K**2).reshape(K,K) * nnz
            self.N_Y = np.random.dirichlet([0.5]*K**2).reshape(K,K) * nnzsum
            #self.N_Y = np.random.poisson(0.1, (K,K)) * N

            if self._is_symmetric:
                self.N_theta_left = self.N_theta_right
                self.N_phi = np.triu(self.N_phi) + np.triu(self.N_phi, 1).T
                self.N_Y = np.triu(self.N_Y) + np.triu(self.N_Y, 1).T


    def _update_gstep_theta(self, idxs):
        ''' Gradient converge for kappa _in (0.5,1] '''
        chi = self._chi_a
        tau = self._tau_a
        kappa = self._kappa_a

        self.gstep_theta[idxs] = chi / ((tau + self._timestep_a[idxs])**kappa)

    def _update_gstep_phi(self):
        chi = self._chi_b
        tau = self._tau_b
        kappa = self._kappa_b

        self.gstep_phi =  chi / ((tau + self._timestep_b)**kappa)

    def _update_gstep_y(self):
        chi = self._chi_b
        tau = self._tau_b
        kappa = self._kappa_b

        self.gstep_y =  chi / ((tau + self._timestep_c)**kappa)


    def _reduce_latent(self):
        theta = self.N_theta_right + self.N_theta_left + np.tile(self.hyper_theta, (self.N_theta_left.shape[0],1))
        self._theta = (theta.T / theta.sum(axis=1)).T

        self._k = self.N_Y + self.hyper_phi[0]
        self._p = (self.N_phi + self.hyper_phi[1] + 1)**-1
        #self._phi = lambda x:sp.stats.nbinom.pmf(x, k, 1-p)
        self._phi = self._k*self._p / (1-self._p)

        #mean = k*p / (1-p)
        #var = k*p / (1-p)**2

        return self._theta, self._phi

    def _reduce_one(self, i, j, xij, update_local=True, update_kernel=True):

        if update_local:
            if self._is_symmetric:
                self.pik = self.pjk = self.N_theta_left[i] + self.hyper_theta
                self.pjk = self.pik
            else:
                self.pik = self.N_theta_left[i] + self.hyper_theta
                self.pjk = self.N_theta_right[j] + self.hyper_theta

        if update_kernel:

            k1 = np.where(np.random.multinomial(1, self.N_theta_left[i]) == 1)[0]
            k2 = np.where(np.random.multinomial(1, self.N_theta_right[j]) == 1)[0]

            btensor = np.outer(self.lbd.T ,self.phi).reshape(self._K, self._K, self._L)
            # Get diagonal
            btensor = np.array([btensor[i:i+self._K,i:i+self._K] for i in range(0,len(btensor),self._K)])
            norm = btensor.sum(0)
            np_count = []
            for b in range(len(btensor)):
                btensor[b] /= norm
                np_count.append(np.random.multinomial(xij, btensor[b].flatten()).reshape(self._K, self._K))

            ksi = np.stack(np_count)


            self.ksi = np.random.multinomial(xij, self.lbd[k1]*self.phi[k2]/norm)
            self.lbd = np.random.dirichlet(0.1 + self.ksi.sum(0) + sum.ksi.sum(1))

            self.pl = self.random.beta(self.ce_eps)
            self.phi = 0
            ###  Unfinished.

            self._kernel = defaultdict2(lambda x:sp.stats.poisson.pmf(x, self.lbd.dot(self.phi).dot(self.lbd.T)))

            if self._hyper_phi == 'auto':
                N = len(self.pik)
                rk = self.hyper_phi[0]
                qk = self.hyper_phi[1]
                pk = 1 / (qk+1)

                # Way 1
                ##print('Gamma %s, %s' % (self.c0*self.r0, 1/(self.c0 - N*np.log(1-pk))))
                a = self.c0_r0
                _pk = 1-pk
                _pk[_pk < 1e-100] = 1e-100
                n=self.N_phi-1
                n[n<0] = 0
                b = 1/(self.c0 - n*np.log(_pk))
                rk = np.random.gamma(a, b)
                #rk = a*b

                c = self.ce_eps + self.N_Y
                d = self.ce_minus_eps + rk*self.N_phi

                #pk = c/(c+d)
                pk = np.random.beta(c, d)
                pk[pk < 1e-100] = 1e-100


                #print('Beta %s, %s' % (self.ce*self.eps + self.N_Y, self.ce*(1-self.eps) + N*rk))
                self.hyper_phi = [rk, (1-pk)/pk]

                # No Residu, in CGS? since p(F, Phi|Z, Y) = q(F,Phi|Z)
                #self._residu = np.array([sp.stats.gamma.pdf(rk, self.c0*self.r0, scale=1/self.c0), sp.stats.beta.pdf(pk, self.ce*self.eps, self.ce*(1-self.eps)) ])

        kernel = self._kernel[xij]

        # debug: Underflow
        kernel[kernel<=1e-300] = 1e-100
        #kernel = ma.masked_invalid(kernel)

        outer_kk = np.log(np.outer(self.pik, self.pjk)) + np.log(kernel) #+ np.log(self._residu).sum()

        return lognormalize(outer_kk.ravel())

    def _optimize_hyper(self):
        n, t = self.hyper_phi
        K = self.N_theta_left.shape[1]

        phi_mean = np.triu(self._phi) if self._is_symmetric else self._phi

        # debug: Underflow
        phi_sum = phi_mean.sum()
        #phi_mean[phi_mean<=1e-300] = 1e-300
        log_phi_sum = np.log(phi_mean).sum()

        K_len = (K*(K-1)/2 +K) if self._is_symmetric else K**2

        t = n * K_len / phi_sum

        # http://bariskurt.com/calculating-the-inverse-of-digamma-function/
        #s = log_phi_sum / K_len + np.log(t)
        #x0 = lambda x: -1/(x + sp.special.digamma(1)) if x<-2.22 else np.exp(x)+0.5

        s = -log_phi_sum / K_len - np.log(K_len/phi_sum)
        x0 = lambda x: (3 - x + np.sqrt((x-3)**2 + 24*x))/(12*x)
        dig = lambda x: np.log(x) - sp.special.digamma(x) -s
        n = float(sp.optimize.minimize(dig, x0(s)).x)

        self.hyper_phi = np.array([n,t])

    def get_mean_phi(self):
        return  self._k*self._p / (1-self._p)

    def get_var_phi(self):
        return self._k*self._p / (1-self._p)**2

    def likelihood(self, theta=None, phi=None):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi


        #_likelihood = defaultdict2(lambda x : sp.stats.poisson.pmf(x, phi))
        _likelihood = defaultdict2(lambda x:sp.stats.nbinom.pmf(x, self._k, 1-self._p))
        qijs = np.array([ theta[i].dot(_likelihood[xij]).dot(theta[j]) for i,j,xij in self.data_test])

        self._likelihood = _likelihood
        #qijs = ma.masked_invalid(qijs)
        return qijs

    def compute_entropy(self, theta=None, phi=None, **kws):
        return self.compute_entropy_t(theta, phi, **kws)

    def compute_entropy_t(self, theta=None, phi=None, **kws):
        if not hasattr(self, 'data_test'):
            return np.nan

        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        #weights = self.data_test[:,2].T
        #ll = sp.stats.poisson.pmf(weights, pij)
        ll = pij

        ll[ll<=1e-300] = 1e-100
        # Log-likelihood
        ll = np.log(ll).sum()
        # Perplexity is 2**H(X).
        return ll

    def compute_elbo(self, theta=None, phi=None, **kws):
        # how to compute elbo for all possible links weights, mean?
        return None

    def compute_roc(self, theta=None, phi=None, treshold=None, **kws):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = np.squeeze(self.data_test[:,2].T)

        #treshold = treshold or 'mean_data'

        if treshold == 'mean_data':
            mean = weights[weights>0].mean()
            std = weights[weights>0].std()
            trsh = int(mean)
            #trsh = int(mean+std)
        elif treshold == 'mean_model':
            mean_, var_ = self.get_mean_phi(), self.get_var_phi()
            mean = mean_.mean()
            std = var_.mean()**0.5
            trsh = int(mean)
            #trsh = int(mean+std)
        else:
            trsh = int(self.expe.get('shift_w',1)) if treshold is None else treshold

        trsh = trsh if trsh>=0 else 1

        self._probas = np.array([ 1 - sum([theta[i].dot(self._likelihood[v]).dot(theta[j]) for v in range(trsh)]) for i,j,_ in self.data_test])

        y_true = weights.astype(bool)*1
        self._y_true = y_true

        fpr, tpr, thresholds = roc_curve(y_true, self._probas)
        roc = auc(fpr, tpr)
        self._eta.append(roc)
        return roc

    def compute_pr(self, *args, **kwargs):
        from sklearn.metrics import average_precision_score
        return average_precision_score(self._y_true, self._probas)

    def mask_probas(self, *args):
        # Copy of compute_roc
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        theta, phi = self._reduce_latent()

        weights = self.data_test[:,2]

        treshold = 'mean_model'
        if treshold == 'mean_data':
            mean = weights[weights>0].mean()
            std = weights[weights>0].std()
        if treshold == 'mean_model':
            mean_, var_ = self.get_mean_phi(), self.get_var_phi()
            mean = mean_.mean()
            std = var_.mean()**0.5

        #trsh = int(mean_w)
        trsh = int(mean+std)

        qijs = self.likelihood(theta, phi)
        y_true = weights.astype(bool)*1
        probas = 1 - sp.stats.poisson.cdf(trsh, qijs)

        return y_true, probas


    def compute_wsim(self, theta=None, phi=None, **kws):
        if 'likelihood' in kws:
            pij = kws['likelihood']
        else:
            if theta is None:
                theta, phi = self._reduce_latent()
            pij = self.likelihood(theta, phi)

        weights = self.data_test[:,2].T
        nnz = self.data_test.shape[0]

        ws = np.array([ theta[i].dot(phi).dot(theta[j]) for i,j,_ in self.data_test])

        # l1 norm
        mean_dist = np.abs(ws - weights).sum() / nnz
        return mean_dist


    def generate(self, N=None, K=None, hyperparams=None, mode='predictive', symmetric=True, **kwargs):
        #self.update_hyper(hyperparams)
        #alpha, gmma, delta = self.get_hyper()

        # predictive
        try: theta, phi = self.get_params()
        except: return self.generate(N, K, hyperparams, 'generative', symmetric)
        K = theta.shape[1]

        raise NotImplementedError

        pij = self.likelihood(theta, phi)
        pij = np.clip(pij, 0, 1)
        Y = sp.stats.bernoulli.rvs(pij)

        return Y

    def fit(self, frontend):
        ''' chunk is the number of row to process in a minibach '''

        self._init(frontend)

        # Init sampling variables
        observed_pt = 0
        mnb_num = 0
        vertex = None

        self.BURNIN = 150
        qij_samples = []
        node_idxs = []
        weights = []
        _qijs_sum = 0
        _qijs_w_sum = 0
        _norm = 0

        self._entropy = self.compute_entropy()
        print( '__init__ Entropy: %f' % self._entropy)
        for _it, obj in enumerate(frontend):

            source, target, weight = obj
            if type(source) is str:
                #print(_it, source, target, weight)
                _set_pos = source
                _vertex = target['vertex']
                _direction = target['direction']
                _scaler = weight
                new_mnb = True

                update_kernel = True
                update_local = True
                burnin = 0
            else:
                i = source
                j = target
                weights.append(weight)
                if direction == 0:
                    node_idxs.append(j)
                else:
                    node_idxs.append(i)

                # Maximization
                qij_samples.append( self._reduce_one(i,j, weight, update_local, update_kernel).reshape(self._len['K'], self._len['K']) )

                observed_pt += 1
                burnin += 1
                update_kernel = False
                update_local = False

            if (new_mnb or burnin % self.BURNIN == 0) and qij_samples:
                qijs = np.asarray(qij_samples)
                ## Update global gradient / Expectation
                ##norm=1
                norm = qijs.shape[0]
                qijs_sum = qijs.sum(0)

                _qijs_sum += qijs_sum
                if set_pos != '0':
                    _qijs_w_sum += np.sum([weights[n]*qijs[n] for n in range(len(weights)) ],0)
                _norm += norm

                gstep_v = self.gstep_theta[vertex]
                gstep_nodes = self.gstep_theta[node_idxs][None].T

                if direction == 0:
                    self.N_theta_left[i] = (1-gstep_v)*self.N_theta_left[i] + gstep_v*scaler*qijs_sum.sum(0) /norm
                    self.N_theta_right[node_idxs] = (1-gstep_nodes)*self.N_theta_right[node_idxs] + gstep_nodes*scaler*qijs.sum(2)
                else:
                    self.N_theta_left[node_idxs] = (1-gstep_nodes)*self.N_theta_left[node_idxs] + gstep_nodes*scaler*qijs.sum(1)
                    self.N_theta_right[j] = (1-gstep_v)*self.N_theta_right[j] + gstep_v*scaler*qijs_sum.sum(1) /norm

                self._timestep_a[vertex] += norm
                self._timestep_a[node_idxs] += 1
                self._update_gstep_theta([vertex]+node_idxs)

                qij_samples.clear()
                node_idxs.clear()
                weights.clear()

                update_local = True


            if new_mnb:
                if vertex is None:
                    # Enter here only once !%!
                    mnb_total = frontend.num_mnb()
                    self.begin_it = time()

                    set_pos = _set_pos
                    vertex = _vertex
                    direction = _direction
                    scaler = _scaler
                    new_mnb = False
                    continue


                self.N_phi = (1 - self.gstep_phi)*self.N_phi + self.gstep_phi * scaler * _qijs_sum /_norm
                if set_pos != '0':
                    self.N_Y = (1 - self.gstep_y)*self.N_Y + self.gstep_y * scaler * _qijs_w_sum /_norm
                    self._timestep_c += _norm
                    self._update_gstep_y()

                    #if self._hyper_phi == 'auto' and mnb_num % 50 == 0:
                    #    self._optimize_hyper()

                self._timestep_b += _norm
                self._update_gstep_phi()


                # Allocate current state variable
                set_pos = _set_pos
                vertex = _vertex
                direction = _direction
                scaler = _scaler

                _qijs_sum = 0
                _qijs_w_sum = 0
                _norm = 0

                new_mnb = False
                mnb_num += 1

                if mnb_num % (self.expe['zeros_set_len']*5) == 0:
                    prop_edge = observed_pt / self._len['nnz']
                    self._observed_pt = observed_pt
                    self.compute_measures()

                    print('.', end='')
                    self.log.info('it %d | prop edge: %.2f | mnb %d/%d, %s, Entropy: %f,  diff: %f' % (_it, prop_edge,
                                                                                                       mnb_num, mnb_total,
                                                                                                       '/'.join((self.expe.model, self.expe.corpus)),
                                                                                                       self._entropy, self.entropy_diff))

                    if self._check_eta():
                        break

                    if self.expe.get('_write'):
                        self.write_current_state(self)
                        if mnb_num % 2000 == 0:
                            self.save(silent=True)
                            sys.stdout.flush()


    def _check_eta(self):

        if self._eta_control > 0:
            self._eta_count -=1
            if self._eta_count == 0:
                if self._eta[-1] - self._eta_control < self._eta_limit:
                    self.log.warning('Exit iteration cause eta criteria met.')
                    return True
                else:
                    self._eta_count = self._eta_count_init
                    self._eta_control = -1
                    self._eta = [self._eta[-1]]
        elif len(self._eta) > 10:
            if self._eta[-1] - self._eta[0] < self._eta_limit:
                self._eta_control = self._eta[-1]
                print('-', end='')
            else:
                self._eta = [self._eta[-1]]






import inspect
from importlib import import_module
import sys
import re
from time import time
import pickle
from copy import deepcopy
import logging

import numpy as np
import scipy as sp

import pymake.io as io


#import sppy


# Todo: rethink sampler and Likelihood class definition.

def mmm(fun):
    # Todo / wrap latent variable routine
    # text vs networks type of data ?
    # * likelihood ?
    # * simlaritie ?
    return fun

class ModelBase():
    """"  Root Class for all the Models.

    * Suited for unserpervised model
    * Virtual methods for the desired propertie of models
    """

    __abstractmethods__ = 'model'

    default_settings = {
        '_write' : False,
        '_csv_typo' : None,
        '_fmt' : None, # unused...
        'iterations' : 3,
        'snapshot_freq': 42,
        'burnin' :  5, # (inverse burnin, last sample to keep
        'thinning' : 1,
    }

    log = logging.getLogger('root')

    def __init__(self, expe=None, frontend=None):
        """ Model Initialization strategy:
            1. self lookup from child initalization
            2. kwargs lookup
            3. default value
        """

        self.expe = expe
        self.frontend = frontend

        # change to semantic -> update value (t+1)
        self.samples = [] # actual sample
        self._samples    = [] # slice to save to avoid writing disk a each iteratoin. (ref format.write_current_state.)

        for k, v in self.default_settings.items():
            self._set_default_settings(k, expe, v)

        # @debug Frontend integratoin !
        # dev a Frontend.get_properties
        if hasattr(self.frontend, 'is_symmetric'):
            self._is_symmetric = self.frontend.is_symmetric()
        if hasattr(self.frontend, 'data_ma'):
            self.mask = self.frontend.data_ma.mask
        #np.fill_diagonal(self.frontend.data_ma, np.ma.masked)

        self._purge_objects = ['frontend', 'data_A', 'data_B']
        if hasattr(self, '_purge'):
            self._purge_objects.extend(self._purge)


    def _set_default_settings(self, key, expe, default):
        if key in expe:
            value = expe[key]
        elif hasattr(self, key):
            value = getattr(self, key)
        else:
            value = default

        return setattr(self, key, value)

    def _init(self, *args, **kwargs):
        ''' Init for fit method.
            Should initialize parmas that depend on the frontend/data.
        '''

        if hasattr(self, '_check_measures'):
            self._check_measures()

        if hasattr(self, '_init_params'):
            self._init_params(*args, **kwargs)

        self.begin_it = time()


    # @Dense mmsb
    def data_iter(self, data=None, randomize=False):
        ''' Iterate over various type of data :
            * ma.array (ignore masked
            * ndarray (todo ?)
            * What format for temporal/chunk/big data... ?
        '''
        return self._data_iter_ma(data, randomize)

    # @Dense mmsb
    def _data_iter_ma(self, data, randomize):
        if data is None:
            data_ma = self.frontend.data_ma
        else:
            data_ma = data

        order = np.arange(data_ma.size).reshape(data_ma.shape)
        masked = order[data_ma.mask]

        if self._is_symmetric:
            tril = np.tril_indices_from(data_ma, -1)
            tril = order[tril]
            masked =  np.append(masked, tril)

        # Remove masked value to the iteration list
        order = np.delete(order, masked)
        # Get the indexes of nodes (i,j) for each observed interactions
        order = list(zip(*np.unravel_index(order, data_ma.shape)))

        if randomize is True:
            np.random.shuffle(order)
        return order

    def getK(self):
        theta, _ = self.get_params()
        return theta.shape[1]

    def getN(self):
        theta, _ = self.get_params()
        return theta.shape[0]

    def _init_params(self, *args, **kwargs):
        pass

    def _check_measures(self):
        if self.expe.get('deactivate_measures'):
            for m in self.expe.get('_csv_typo', '').split():
                if not hasattr(self, m):
                    setattr(self, m, None)

    def compute_measures(self):
        ''' Compute measure as model attributes.
            begin_it: is the time of the begining of the iteration.
        '''

        if self.expe.get('deactivate_measures'):
            return

        if hasattr(self, 'begin_it'):
            self.time_it = time() - self.begin_it

        params = self._reduce_latent()
        ll = self.likelihood(*params)
        kws = {'likelihood':ll}

        for meas in self.expe._csv_typo.split():

            if meas.lstrip('_') == 'entropy':
                old_entropy = getattr(self, '_entropy', -np.inf)
                _meas = self.compute_entropy(*params, **kws)
                self.entropy_diff = _meas - old_entropy
            elif hasattr(self, 'compute_'+meas.lstrip('_')):
                _meas = getattr(self, 'compute_'+meas.lstrip('_'))(*params, **kws)
            else:
                # Assume already computed
                getattr(self, meas) # raise exception if not here.
                continue

            setattr(self, meas, _meas)

        return



    @mmm #frontend
    def likelihood(self, theta=None, phi=None):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi
        likelihood = theta.dot(phi).dot(theta.T)
        return likelihood

    @mmm
    def similarity_matrix(self, theta=None, phi=None, sim='cos'):
        if theta is None:
            theta = self._theta
        if phi is None:
            phi = self._phi

        features = theta
        if sim in  ('dot', 'latent'):
            sim = np.dot(features, features.T)
        elif sim == 'cos':
            norm = np.linalg.norm(features, axis=1)
            sim = np.dot(features, features.T)/norm/norm.T
        elif sim in  ('model', 'natural'):
            sim = features.dot(phi).dot(features.T)
        else:
            self.log.error('Similaririty metric unknow: %s' % sim)
            sim = None

        if hasattr(self, 'normalization_fun'):
            sim = self.normalization_fun(sim)
        return sim


    def get_params(self):
        if hasattr(self, '_theta') and hasattr(self, '_phi'):
            return np.asarray(self._theta), np.asarray(self._phi)
        else:
            return self._reduce_latent()

    def _reduce_latent(self):
        ''' Estimate global parameters of a model '''
        raise NotImplementedError


    def update_hyper(self):
        self.log.warning('No method to update hyperparams..')
        return

    def get_hyper(self):
        self.log.error('no method to get hyperparams')
        return

    def save(self, silent=False):
        to_remove = []
        for k, v in self.__dict__.items():
            if hasattr(v, 'func_name') and v.func_name == '<lambda>':
                to_remove.append(k)
            if str(v).find('<lambda>') >= 0:
                # python3 hook, nothing better ?
                to_remove.append(k)
            #elif type(k) is defaultdict:
            #    setattr(self.model, k, dict(v))

        if to_remove or self._has_purge():
            model = deepcopy(self)
            model.purge()

            for k in to_remove:
                try:
                    delattr(model, k)
                except Exception as e:
                    self.log.debug('Cant delete object during model purging: %s' % e)
        else:
            model = self
            if hasattr(self, 'write_current_state'):
                delattr(self, 'write_current_state')


        fn = self.expe['_output_path']
        if not silent:
            self.log.info('Snapshotting Model : %s' % fn)
        else:
            sys.stdout.write('+')

        io.save(fn, model, silent=True)


    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            try:
                setattr(result, k, deepcopy(v, memo))
            except Exception as e:
                self.log.debug('can\'t copy %s : %s. Passing on: %s' % (k, v, e))
                continue
        return result

    #@dense mmsb
    def get_mask(self):
        return self.mask

    #@dense mmsb
    def mask_probas(self, data):
        mask = self.get_mask()
        y_test = data[mask]
        p_ji = self.likelihood(*self._reduce_latent())
        probas = p_ji[mask]
        return y_test, probas

    #@dense mmsb
    def compute_roc(self):
        from sklearn.metrics import roc_curve, auc, precision_recall_curve

        data = self.frontend.data

        y_true, probas = self.mask_probas(data)
        fpr, tpr, thresholds = roc_curve(y_true, probas)

        roc = auc(fpr, tpr)
        return roc


    #@dense mmsb
    @mmm
    def predictMask(self, data, mask=True):
        self.log.info('Reducing latent variables...')

        if mask is True:
            masked = self.get_mask()
        else:
            masked = mask

        ### @Debug Ignore the Diagonnal when predicting.
        np.fill_diagonal(masked, False)

        ground_truth = data[masked]

        p_ji = self.likelihood(*self.get_params())
        prediction = p_ji[masked]
        prediction = sp.stats.bernoulli.rvs( prediction )
        #prediction[prediction >= 0.5 ] = 1
        #prediction[prediction < 0.5 ] = 0

        ### Computing Precision
        test_size = float(ground_truth.size)
        good_1 = ((prediction + ground_truth) == 2).sum()
        precision = good_1 / float(prediction.sum())
        rappel = good_1 / float(ground_truth.sum())
        g_precision = (prediction == ground_truth).sum() / test_size
        mask_density = ground_truth.sum() / test_size

        ### Finding Communities
        if hasattr(self, 'communities_analysis'):
            self.log.info('Finding Communities...')
            communities = self.communities_analysis(data)
            K = self.K
        else:
            communities = None
            K = self.expe.get('K')

        res = {'Precision': precision,
               'Recall': rappel,
               'g_precision': g_precision,
               'mask_density': mask_density,
               'clustering':communities,
               'K': K
              }
        return res


    def fit(self, *args, **kwargs):
        ''' A core method.

            Template
            --------

            # cant take argument, they will be passed to
            # _init_params that you can safely  overwrite.
            #
            self._init()

            for _it in range(self.expe.iterations):
                # core process

                if self.expe.get('_write'):
                    self.write_current_state(self)

                    # In addition to that, model is automatically
                    # saved at the end of a script if the model
                    # is configured ie (called to load_model())
                    #
                    if _it > 0 and _it % self.snapshot_freq == 0:
                        self.save(silent=True)
                        sys.stdout.flush()
        '''
        raise NotImplementedError

    def transform(self, *args, **kwargs):
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        raise NotImplementedError

    # Search ?


    def generate(self):
        raise NotImplementedError
    def get_clusters(self):
        raise NotImplementedError

    def _has_purge(self):
        return any([getattr(self, o, None) for o in self._purge_objects])

    def purge(self):
        for obj in self._purge_objects:
            if hasattr(self, obj):
                delattr(self, obj)



class ModelSkl(ModelBase):

    '''
    Notes
    -----
    Model class need to be serialisable. Module object are not serialisable.
    Avoid keeping it in the self object.
    '''

    def __init__(self, expe, frontend=None):
        super(ModelSkl, self).__init__(expe, frontend)

        # Load Sklearn Model
        if not hasattr(self, 'module'):
            self.log.error('ModelSkl base class need a {module} name attribute. Exiting.')
            exit(42)

        _module, _model = self._mm_from_str(self.module)
        self._spec = self._spec_from_expe(_model)

        # Init Sklearn model
        self.model = _model(**self._spec)


    @staticmethod
    def _mm_from_str(module):
        _module = module.split('.')
        _module, model_name = '.'.join(_module[:-1]), _module[-1]
        module = import_module(_module)
        _model = getattr(module, model_name)
        return module, _model

    def _spec_from_expe(self, _model):
        ''' Set Sklearn parameters. '''

        model_params = list(inspect.signature(_model).parameters)
        spec = dict()
        spec_map = getattr(self, 'spec_map', {})
        default_spec = getattr(self, '_default_spec', {})
        for k in model_params:
            if k in list(self.expe)+list(spec_map):
                _k = spec_map.get(k, k)
                if _k in self.expe:
                    spec[k] = self.expe[_k]
            elif k in default_spec:
                spec[k] = default_spec[k]

        return spec

    def __getattr__(self, attr):
        ''' Propagate sklearn attribute.

            Notes
            -----
            __getatrr__ is call only if the attr doesn't exist...
        '''

        if not 'model' in self.__dict__:
            raise AttributeError

        attr = attr.partition('__hack_me_')[-1]
        return getattr(self.model, attr)

    def fit(self, *args, **kwargs):
        fun =  self.__hack_me_fit
        self.log.info("Fitting `%s' model with spec: %s" % (type(self), str(self._spec)))
        return fun(*args, **kwargs)

    def transform(self, *args, **kwargs):
        fun =  self.__hack_me_transform
        data = fun(*args, **kwargs)

        if hasattr(self, 'post_transform'):
            for module in self.post_transform:
                _m, _model = self._mm_from_str(module)
                spec = self._spec_from_expe(_model)
                model = _model(**spec)
                data = model.fit_transform(data)

        return data

    def fit_transform(self, *args, **kwargs):
        fun =  self.__hack_me_fit_transform
        return fun(*args, **kwargs)

    # @Obsolete ?
    def predict(self, *args, **kwargs):
        fun =  self.__hack_me_predict
        return fun(*args, **kwargs)

    # get_params()
    # set_params()


class GibbsSampler(ModelBase):
    ''' Implemented method, except fit (other?) concerns MMM type models :
        * LDA like
        * MMSB like

        but for other (e.g IBP based), method has to be has to be overloaded...
        -> use a decorator @mmm to get latent variable ...
    '''
    __abstractmethods__ = 'model'
    def __init__(self, expe, frontend):
        super(GibbsSampler, self).__init__(expe, frontend)

    @mmm
    def compute_measures(self, begin_it=0):

        if self.expe.get('deactivate_measures'):
            return

        self._entropy = self.compute_entropy()

        if hasattr(self, 'compute_entropy_t'):
            self._entropy_t = self.compute_entropy_t()
        else:
            self._entropy_t = np.nan

        if '_roc' in self.expe._csv_typo.split():
            self._roc = self.compute_roc()

        self._alpha = np.nan
        self._gmma= np.nan
        self.alpha_mean = np.nan
        self.delta_mean = np.nan
        self.alpha_var = np.nan
        self.delta_var= np.nan

        self.time_it = time() - begin_it

    def fit(self, *args, **kwargs):

        self._init()

        self._entropy = self.compute_entropy()
        print( '__init__ Entropy: %f' % self._entropy)

        for _it in range(self.iterations):
            self._iteration = _it

            ### Sampling
            begin_it = time()
            self.sample()

            if _it >= self.iterations - self.burnin:
                if _it % self.thinning == 0:
                    self.samples.append([self._theta, self._phi])

            self.compute_measures(begin_it)
            print('.', end='')
            self.log.info('iteration %d, %s, Entropy: %f \t\t K=%d  alpha: %f gamma: %f' % (_it, '/'.join((self.expe.model, self.expe.corpus)),
                                                                                    self._entropy, self._K,self._alpha, self._gmma))
            if self.expe.get('_write'):
                self.write_current_state(self)
                if _it > 0 and _it % self.snapshot_freq == 0:
                    self.save(silent=True)
                    sys.stdout.flush()

        ### Clean Things
        if not self.samples:
            self.samples.append([self._theta, self._phi])
        self._reduce_latent()
        self.samples = None # free space
        return


    @mmm
    def update_hyper(self, hyper):
        if hyper is None:
            return
        elif isinstance(type(hyper), (tuple, list)):
            alpha = hyper[0]
            gmma = hyper[1]
            delta = hyper[2]
        else:
            delta = hyper.get('delta')
            alpha = hyper.get('alpha')
            gmma = hyper.get('gmma')

        if delta:
            self._delta = delta
        if alpha:
            self._alpha = alpha
        if gmma:
            self._gmma = gmma


    # keep only the most representative dimension (number of topics) in the samples
    @mmm
    def _reduce_latent(self):
        theta, phi = list(map(list, zip(*self.samples)))
        ks = [ mat.shape[1] for mat in theta]
        bn = np.bincount(ks)
        k_win = np.argmax(bn)
        self.log.debug('K selected: %d' % k_win)

        ind_rm = []
        [ind_rm.append(i) for i, v in enumerate(theta) if v.shape[1] != k_win]
        for i in sorted(ind_rm, reverse=True):
            theta.pop(i)
            phi.pop(i)

        self.log.debug('Samples Selected: %d over %s' % (len(theta), len(theta)+len(ind_rm) ))

        self._theta = np.mean(theta, 0)
        self._phi = np.mean(phi, 0)
        self.K = self._theta.shape[1]
        return self._theta, self._phi



class SVB(ModelBase):

    '''online EM/SVB

        Note: this obsolete base class is suited for dense masked numpy array.
              Base class for dense matrix svb...
    '''

    __abstractmethods__ = 'model'

    def __init__(self, expe, frontend):
        super(SVB, self).__init__(expe, frontend)

        self.entropy_diff = 0

    def _update_chunk_nnz(self, groups):
        _nnz = []
        for g in groups:
            if self._is_symmetric:
                count = 2* len(g)
                #count =  len(g)
            else:
                count = len(g)
            _nnz.append(count)

        self._nnz_vector = _nnz

    def _get_chunk(self):
        chunk = self.expe.get('chunk', 100)

        if isinstance(chunk, (int, float)):
            return chunk
        elif isinstance(chunk, str):
            pass
        else:
            raise TypeError('Unknown chunk type: %s' % type(chunk))

        mean_nodes = np.mean(self._len['dims'])
        try:
            chunk_mode, ratio = chunk.split('_')
        except ValueError as e:
            return float(chunk)

        if chunk_mode == 'adaptative':
            if '.' in ratio:
                ratio = float(ratio)
            else:
                ratio = int(ratio)

            chunk = ratio * mean_nodes
        else:
            raise ValueError('Unknown chunk mode: %s' % chunk_mode)

        return chunk


    def fit(self, *args, **kwargs):
        ''' chunk is the number of row to threat in a minibach '''

        self._init()

        data_ma = self.frontend.data_ma
        groups = self.data_iter(data_ma, randomize=True)
        chunk_groups = np.array_split(groups, self.chunk_len)
        self._update_chunk_nnz(chunk_groups)

        self._entropy = self.compute_entropy()
        print( '__init__ Entropy: %f' % self._entropy)
        for _id_mnb, minibatch in enumerate(chunk_groups):

            self.mnb_size = self._nnz_vector[_id_mnb]

            begin_it = time()
            self.sample(minibatch)

            self.compute_measures(begin_it)
            print('.', end='')
            self.log.info('Minibatch %d/%d, %s, Entropy: %f,  diff: %f' % (_id_mnb+1, self.chunk_len, '/'.join((self.expe.model, self.expe.corpus)),
                                                                        self._entropy, self.entropy_diff))
            if self.expe.get('_write'):
                self.write_current_state(self)
                if _id_mnb > 0 and _id_mnb % self.snapshot_freq == 0:
                    self.save(silent=True)
                    sys.stdout.flush()

    def sample(self, minibatch):
        '''
        Notes
        -----
        * self.iterations is actually the burnin phase of SVB.
        '''

        # not good
        #self._timestep_a = 0

        # self.iterations is actually the burnin phase of SVB.
        for _it in range(self.iterations+1):
            self.log.debug('init timestep A ?')
            #self._timestep_a = 0 # test, not good
            self._iteration = _it
            burnin = (_it < self.iterations)
            np.random.shuffle(minibatch)
            for _id_token, iter in enumerate(minibatch):
                self._id_token = _id_token
                self._xij = self.frontend.data_ma[tuple(iter)]
                self.maximization(iter)
                self.expectation(iter, burnin=burnin)


            if self._iteration != self.iterations-1 and self.expe.get('_verbose', 20) < 20:
                self.compute_measures()
                self.log.debug('it %d,  ELBO: %f, elbo diff: %f \t K=%d' % (_it, self._entropy, self.entropy_diff, self._K))
                if self.expe.get('_write'):
                    self.write_current_state(self)


        # Update global parameters after each "minibatech of links"
        # @debug if not, is the time_step evolve differently for N_Phi and N_Theta.
        minibashed_finished = True
        if minibashed_finished:
            self._purge_minibatch()

        #if self.elbo_diff < self.limit_elbo_diff:
        #    print('ELBO get stuck during data iteration : Sampling useless, return ?!')






import sys
import setuptools
from codecs import open
from os import path

#__version__ = subprocess.check_output(["git", "describe"]).strip()
__version__ = '0.42.1'

try:
    from Cython.Build import cythonize
    #from distutils.extension import Extension
except ImportError:
    CYTHON = False
else:
    CYTHON = 'bdist_wheel' not in sys.argv


# Requirements
_here = path.abspath(path.dirname(__file__))
with open(path.join(_here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
install_requires = [x.strip() for x in all_reqs if not (x.strip().startswith(('#', '//')) or x.strip().endswith('?'))]
install_requires = list(filter(None, install_requires))

# Packages
packages = setuptools.find_packages()
packages += ['repo/ml'] +  ['repo/ml/'+p for p in setuptools.find_packages('repo/ml/')]


setuptools.setup(
    name='pmk',
    version=__version__,
    author='dtrckd',
    author_email='dtrckd@gmail.com',
    description='An experiment control system for reproducible research.',
    url='https://github.com/dtrckd/pymake',
    license='GPL',
    install_requires=install_requires,
    entry_points={
        'console_scripts': ['pmk=pymake.zymake:main'],
    },
    packages=packages,
    package_dir={'ml' : 'repo/ml'},
    package_data={'pymake' : ['pmk.cfg', 'template/*.template']},
    include_package_data=True,
    keywords=['pymake', 'learning', 'model'],
    classifiers=[
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
    ]

    ### Trying to cythonize the project (could help create a binary)
    #ext_modules=cythonize(
    #    "bhp/*.py", "bhp/**/*.py",
    #    #exclude=[
    #    #    'tatsu/__init__.py',
    #    #    'tatsu/codegen/__init__.py',
    #    #    'tatsu/test/*.py'
    #    #]
    #),

)




import os
from collections import defaultdict
from string import Template
import shelve


# Global settings
__default_config = defaultdict(lambda: '', dict(project_data = os.path.expanduser('data/'),
                                                project_figs = os.path.expanduser('data/plot/figs/'),
                                                # @debug repo access ??
                                                default_spec = 'pymake.spec',
                                                default_script = 'pymake.script',
                                                default_model = 'pymake.model',
                                                default_corpus = '?')
                              )

class PmkTemplate(Template):
    delimiter = '$$'
    #idpattern = r'[a-z][_a-z0-9]*'


def parse_file_conf(fn, sep=':', comments=('#','%')):
    with open(fn) as f:
        parameters = f.read()
    parameters = filter(None, parameters.split('\n'))
    parameters = dict((p[0].strip(), p[1].strip()) for p in (t.strip().split(sep) for t in parameters if not t.strip().startswith(comments)))
    for k, v in parameters.items():
        if  '.' in v:
            try:
                parameters[k] = float(v)
            except:
                pass
        else:
            try:
                parameters[k] = int(v)
            except:
                pass
    return parameters

def reset_pymake_settings(settings, default_config=__default_config, cfg_name='pmk.cfg', db_name='.pmk-db'):
    _settings = default_config.copy()
    _settings.update(settings)
    #ctnt = '\n'.join(('{0} = {1}'.format(k,v) for k,v in _settings.items()))
    cwd = os.path.dirname(__file__)
    with open(os.path.join(cwd, '..', 'template', '%s.template'%(cfg_name))) as _f:
        template = PmkTemplate(_f.read())
        ctnt = template.substitute(_settings)

    try:
        db = shelve.open(os.path.join(os.getcwd(), db_name))
        dir = db['PWD']
        db.close()
    except Exception as e:
        print("Bootstrap warning (%s) => PWD path not initialized ? key: %s" % (e, key))
        dir = os.getenv('PWD')

    cfg_file = os.path.join(dir, cfg_name)
    with open(cfg_file, 'wb') as _f:
        return _f.write(ctnt.encode('utf8'))

def get_pymake_settings(key=None, default_config=__default_config, cfg_name='pmk.cfg', db_name='.pmk-db'):
    try:
        db = shelve.open(os.path.join(os.getcwd(), db_name))
        dir = db['PWD']
        db.close()
    except Exception as e:
        print("Bootstrap warning (%s) => PWD path not initialized ? key: %s" % (e, key))
        dir = os.getenv('PWD')

    cfg_file = os.path.join(dir, cfg_name)

    if not os.path.isfile(cfg_file):
        cfg_file = os.path.join(os.path.expanduser('~') ,'.pymake', cfg_name)
        if not os.path.isfile(cfg_file):
            dir_cfg = make_path(cfg_file)
            #ctnt = '\n'.join(('{0} = {1}'.format(k,v) for k,v in  default_config.items()))
            ctnt = ''
            with open(cfg_file, 'wb') as _f:
                _f.write(ctnt.encode('utf8'))

    config = parse_file_conf(cfg_file, sep='=')

    for k in list(config):
        v = config[k]
        if '~' in v:
            config[k] = os.path.expanduser(v)
        elif v and (v.startswith(('./', '../')) or not v.startswith('/')):
            fpath = os.path.join(dir, v)
            if os.path.exists(fpath):
                config[k] = fpath

    if not key:
        settings = config
    elif key.startswith('_'):
        res = []
        for k in ['default'+key, 'contrib'+key]:
            res += config.get(k, default_config[k]).split(',')
        settings =  [e for e in map(str.strip, res) if e]
    else:
        settings = config.get(key, default_config[key])

    return settings



# -*- coding: utf-8 -*-

import numpy as np
from numpy import ma
import scipy as sp
import networkx as nx
from .utils import nxG

import logging
lgg = logging.getLogger('root')

##########################
### Stochastic Process
##########################

def lognormalize(x):
    return np.exp(x - np.logaddexp.reduce(x))

def categorical(params):
    return np.where(np.random.multinomial(1, params) == 1)[0]

def bernoulli(param, size=1):
    return np.random.binomial(1, param, size=size)

### Power law distribution generator
def random_powerlaw(alpha, x_min, size=1):
    ### Discrete
    alpha = float(alpha)
    u = np.random.random(size)
    x = (x_min-0.5)*(1-u)**(-1/(alpha-1))+0.5
    return np.floor(x)

### A stick breakink process, truncated at K components.
def gem(gmma, K):
    sb = np.empty(K)
    cut = np.random.beta(1, gmma, size=K)
    for k in range(K):
        sb[k] = cut[k] * cut[0:k].prod()
    return sb

##########################
### Means and Norms
##########################

### Weighted  means
def wmean(a, w, mean='geometric'):
    if mean == 'geometric':
        kernel = lambda x : np.log(x)
        out = lambda x : np.exp(x)
    elif mean == 'arithmetic':
        kernel = lambda x : x
        out = lambda x : x
    elif mean == 'harmonic':
        num = np.sum(w)
        denom = np.sum(np.asarray(w) / np.asarray(a))
        return num / denom
    else:
        raise NotImplementedError('Mean Unknwow: %s' % mean)

    num = np.sum(np.asarray(w) * kernel(np.asarray(a)))
    denom = np.sum(np.asarray(w))
    return out(num / denom)

##########################
### Matrix/Image Operation
##########################
from scipy import ndimage

def draw_square(mat, value, topleft, l, L, w=0):
    tl = topleft

    # Vertical draw
    mat[tl[0]:tl[0]+l, tl[1]:tl[1]+w] = value
    mat[tl[0]:tl[0]+l, tl[1]+L-w:tl[1]+L] = value
    # Horizontal draw
    mat[tl[0]:tl[0]+w, tl[1]:tl[1]+L] = value
    mat[tl[0]+l-w:tl[0]+l, tl[1]:tl[1]+L] = value
    return mat

def dilate(y, size=1):
    dim = y.ndim
    mask = ndimage.generate_binary_structure(dim, dim)
    if size > 1:
        for i in range(1, size):
            mask = np.vstack((mask, mask[-1,:]))
            mask = np.column_stack((mask, mask[:, -1]))
    y_f = ndimage.binary_dilation(y, structure=mask).astype(y.dtype)
    return y_f


##########################
### Array routine  Operation
##########################
from collections import Counter

def sorted_perm(a, label=None, reverse=False):
    """ return sorted $a and the induced permutation.
        Inplace operation """
    # np.asarray applied this tuple lead to error, if label is string
    # because a should be used as elementwise comparison
    if label is None:
        label = np.arange(a.shape[0])
    hist, label = zip(*sorted(zip(a, label), reverse=reverse))
    hist = np.asarray(hist)
    label = np.asarray(label)
    return hist, label

def degree_hist_to_list(d, dc):
    degree = np.repeat(np.round(d).astype(int), np.round(dc).astype(int))
    return degree


def clusters_hist(clusters, labels=None, remove_empty=True):
    """ return non empty clusters histogramm sorted.

        parameters
        ---------
        clusters: np.array
            array of clusters membership of data.

        returns
        -------
        hist: np.array
            count of element by clusters (decrasing hist)
        label: np.array
            label of the cluster aligned with hist
    """
    block_hist = np.bincount(clusters)
    if labels is None:
        labels = range(len(block_hist))

    hist, labels = sorted_perm(block_hist, labels, reverse=True)

    if remove_empty is True:
        null_classes = (hist == 0).sum()
        if null_classes > 0:
            hist = hist[:-null_classes]; labels = labels[:-null_classes]

    return hist, labels

def adj_to_degree(y):
    # @debug: dont' call nxG or do a native integration !

    # To convert normalized degrees to raw degrees
    #ba_c = {k:int(v*(len(ba_g)-1)) for k,v in ba_c.iteritems()}
    G = nxG(y)
    #degree = sorted(dict(nx.degree(G)).values(), reverse=True)

    #ba_c = nx.degree_centrality(G)
    return dict(nx.degree(G))

def degree_hist(_degree, filter_zeros=False):
    if isinstance(_degree, np.ndarray) and _degree.ndim == 2 :
        degree = list(dict(adj_to_degree(_degree)).values())
    elif isinstance(_degree, (list, np.ndarray)):
        degree = _degree
    else:
        # networkx
        degree = list(dict(_degree).values())

    max_c = np.max(degree)
    d = np.arange(max_c+1)
    dc = np.bincount(degree, minlength=max_c+1)

    if len(d) == 0:
        return [], []

    if dc[0] > 0:
        lgg.debug('%d unconnected vertex' % dc[0])
    d = d[1:]
    dc = dc[1:]


    if filter_zeros is True:
        #d, dc = zip(*filter(lambda x:x[1] != 0, zip(d, dc)))
        nzv = (dc != 0)
        d = d[nzv]
        dc = dc[nzv]

    return d, dc

def random_degree(Y, params=None):
    _X = []
    _Y = []
    N = Y[0].shape[0]
    nb_uniq_degree = []
    dc_list = []
    for y in Y:
        ba_c = adj_to_degree(y)
        d, dc = degree_hist(ba_c)
        nb_uniq_degree.append(len(dc))
        dc_list.append(dc)
    dc_mat = ma.array(np.empty((N, max(nb_uniq_degree))), mask=True)
    for i, degrees in enumerate(dc_list):
        size = nb_uniq_degree[i]
        dc_mat[i, :size] = degrees

    y = dc_mat.mean(0)
    yerr = dc_mat.std(0)
    # 0 are filtered out in degree_hist
    return np.arange(1, len(y)+1),  np.round(y), yerr

def reorder_mat(y, clusters, labels=False, reverse=True):
    """Reorder the  matrix according the clusters membership
        @Debug: square matrix
    """
    assert(y.shape[0] == y.shape[1] == len(clusters))
    if reverse is True:
        hist, label = clusters_hist(clusters)
        sorted_clusters = np.empty_like(clusters)
        for i, k in enumerate(label):
            if i != k:
                sorted_clusters[clusters == k] = i
    else:
        sorted_clusters = clusters

    N = y.shape[0]
    nodelist = [k[0] for k in sorted(zip(range(N), sorted_clusters),
                                     key=lambda k: k[1])]

    y_r = y[nodelist, :][:, nodelist]
    if labels is True:
        return y_r, nodelist
    else:
        return y_r

def shiftpos(arr, fr, to, axis=0):
    """ Move element In-Place, shifting backward (or forward) others """
    if fr == to: return
    x = arr.T if axis == 1 else arr
    tmp = x[fr].copy()
    if fr > to:
        x[to+1:fr+1] = x[to:fr]
    else:
        x[fr:to] = x[fr+1:to+1]
    x[to] = tmp


##########################
### Colors Operation
##########################
import math

def floatRgb(mag, cmin, cmax):
	""" Return a tuple of floats between 0 and 1 for the red, green and
		blue amplitudes.
	"""

	try:
		# normalize to [0,1]
		x = float(mag-cmin)/float(cmax-cmin)
	except:
		# cmax = cmin
		x = 0.5
	blue = min((max((4*(0.75-x), 0.)), 1.))
	red  = min((max((4*(x-0.25), 0.)), 1.))
	green= min((max((4*math.fabs(x-0.5)-1., 0.)), 1.))
	return (red, green, blue)

def strRgb(mag, cmin, cmax):
   """ Return a tuple of strings to be used in Tk plots.
   """

   red, green, blue = floatRgb(mag, cmin, cmax)
   return "#%02x%02x%02x" % (red*255, green*255, blue*255)

def rgb(mag, cmin, cmax):
   """ Return a tuple of integers to be used in AWT/Java plots.
   """

   red, green, blue = floatRgb(mag, cmin, cmax)
   return (int(red*255), int(green*255), int(blue*255))

def htmlRgb(mag, cmin, cmax):
   """ Return a tuple of strings to be used in HTML documents.
   """
   return "#%02x%02x%02x"%rgb(mag, cmin, cmax)

